"""Implementations for Sprite class.
"""

from typing import Optional

from apysc._display import graphics
from apysc._display.child_interface import ChildInterface
from apysc._display.display_object import DisplayObject
from apysc._type.revert_interface import RevertInterface


class Sprite(DisplayObject, ChildInterface, RevertInterface):
    """
    Basic display object that can be parent.

    References
    ----------
    - Sprite document
        - https://simon-ritchie.github.io/apysc/sprite.html
    """

    graphics: 'graphics.Graphics'

    def __init__(
            self, *,
            variable_name: Optional[str] = None) -> None:
        """
        Basic display object that can be parent.

        Parameters
        ----------
        variable_name : str or None, default None
            Variable name of this instance. This will be used to
            js expression. It is not necessary to specify any
            string except when Sprite subclass will be instantiated.

        References
        ----------
        - Sprite document
            - https://simon-ritchie.github.io/apysc/sprite.html

        Examples
        --------
        >>> import apysc as ap
        >>> stage: ap.Stage = ap.Stage()
        >>> sprite_1: ap.Sprite = ap.Sprite()
        >>> # Create the sprite child rectangle
        >>> sprite_1.graphics.begin_fill(color='#0af')
        >>> rect: ap.Rectangle = sprite_1.graphics.draw_rect(
        ...     x=50, y=50, width=50, height=50)
        >>> sprite_1.graphics.contains(rect)
        Boolean(True)

        >>> # Move the created rectangle to the other sprite
        >>> sprite_2: ap.Sprite = ap.Sprite()
        >>> sprite_2.add_child(rect)
        >>> sprite_1.graphics.contains(rect)
        Boolean(False)
        >>> sprite_2.contains(rect)
        Boolean(True)

        >>> # Move the sprite container
        >>> sprite_2.x = ap.Int(50)
        >>> sprite_2.x
        Int(50)
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='__init__', locals_=locals(),
                module_name=__name__, class_=Sprite):
            from apysc._expression import expression_variables_util
            from apysc._expression import var_names
            stage: ap.Stage = ap.get_stage()
            if variable_name is None:
                variable_name = expression_variables_util.\
                    get_next_variable_name(type_name=var_names.SPRITE)
            self._children = ap.Array([])
            super(Sprite, self).__init__(variable_name=variable_name)
            self._append_constructor_expression()
            self.graphics = graphics.Graphics(parent=self)
            stage.add_child(child=self)
            self._set_overflow_visible_setting()

    def _append_constructor_expression(self) -> bool:
        """
        Append Sprite constructor expression.

        Notes
        -----
        Expression not to be added if instance is Sprite subclass.

        Returns
        -------
        appended : bool
            If expression appended, then True will be set.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self._append_constructor_expression,
                locals_=locals(),
                module_name=__name__, class_=Sprite):
            from apysc._display.stage import get_stage_variable_name
            from apysc._type import type_util
            is_same_class_instance: bool = type_util.is_same_class_instance(
                class_=Sprite, instance=self)
            if not is_same_class_instance:
                return False
            stage_variable_name: str = get_stage_variable_name()
            expression: str = (
                f'\nvar {self.variable_name} = {stage_variable_name}.nested();'
            )
            ap.append_js_expression(expression=expression)
            return True

    def _make_snapshot(self, *, snapshot_name: str) -> None:
        """
        Make values' snapshot.

        Parameters
        ----------
        snapshot_name : str
            Target snapshot name.
        """
        if self._snapshot_exists(snapshot_name=snapshot_name):
            return
        self.graphics._run_all_make_snapshot_methods(
            snapshot_name=snapshot_name)

    def _revert(self, *, snapshot_name: str) -> None:
        """
        Revert values if snapshot exists.

        Parameters
        ----------
        snapshot_name : str
            Target snapshot name.
        """
        if not self._snapshot_exists(snapshot_name=snapshot_name):
            return
        self.graphics._run_all_revert_methods(snapshot_name=snapshot_name)

    def __repr__(self) -> str:
        """
        Get a string representation of this instance (for the sake of
        debugging).

        Returns
        -------
        repr_str : str
            Type name and variable name will be set
            (e.g., `Sprite('<variable_name>')`).
        """
        repr_str: str = f"Sprite('{self.variable_name}')"
        return repr_str



"""Class implementation for the animation_pause interface.
"""

from apysc._type.variable_name_interface import VariableNameInterface


class AnimationPauseInterface(VariableNameInterface):

    def animation_pause(self) -> None:
        """
        Stop the all animations.

        References
        ----------
        - animation_pause and animation_play interfaces document
            - https://bit.ly/3m2Xh8Y
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self.animation_pause, locals_=locals(),
                module_name=__name__, class_=AnimationPauseInterface):
            expression: str = (
                f'{self.variable_name}.timeline().pause();'
            )
            ap.append_js_expression(expression=expression)



"""Class implementation for the scale_x_from_point interfaces.
"""

from typing import Any
from typing import Dict

from apysc._animation.animation_scale_x_from_point_interface import \
    AnimationScaleXFromPointInterface
from apysc._type.dictionary import Dictionary
from apysc._type.int import Int
from apysc._type.number import Number
from apysc._type.revert_interface import RevertInterface


class ScaleXFromPointInterface(
        AnimationScaleXFromPointInterface, RevertInterface):

    _scale_x_from_point: Dictionary[str, Number]

    def _initialize_scale_x_from_point_if_not_initialized(self) -> None:
        """
        Initialize the `_scale_x_from_point` attribute if it hasn't been
        initialized yet.
        """
        if hasattr(self, '_scale_x_from_point'):
            return
        self._scale_x_from_point = Dictionary({})

    def get_scale_x_from_point(self, x: Int) -> Number:
        """
        Get a scale-x value from the given x-coordinate.

        Parameters
        ----------
        x : Int
            X-coordinate.

        Returns
        -------
        scale_x : Number
            Scale-x value from the given x-coordinate.

        References
        ----------
        - GraphicsBase scale_from_point interfaces document
            - https://bit.ly/3xRBhlw
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self.get_scale_x_from_point, locals_=locals(),
                module_name=__name__, class_=ScaleXFromPointInterface):
            from apysc._display import scale_interface_helper
            from apysc._type.expression_string import ExpressionString
            from apysc._validation import number_validation
            number_validation.validate_integer(integer=x)
            self._initialize_scale_x_from_point_if_not_initialized()
            default_val: ap.Number = ap.Number(1.0)
            key_exp_str: ExpressionString = scale_interface_helper.\
                get_coordinate_key_for_expression(coordinate=int(x._value))
            scale_x: ap.Number = self._scale_x_from_point.get(
                key=key_exp_str, default=default_val)
            return scale_x

    def set_scale_x_from_point(self, scale_x: Number, x: Int) -> None:
        """
        Update a scale-x value from the given x-coordinate.

        Parameters
        ----------
        scale_x : Number
            Scale-x value to set.
        x : Int
            X-coordinate.

        References
        ----------
        - GraphicsBase scale_from_point interfaces document
            - https://bit.ly/3xRBhlw
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self.set_scale_x_from_point, locals_=locals(),
                module_name=__name__, class_=ScaleXFromPointInterface):
            from apysc._display import scale_interface_helper
            from apysc._type.expression_string import ExpressionString
            from apysc._validation import number_validation
            number_validation.validate_num(num=scale_x)
            number_validation.validate_integer(integer=x)
            self._initialize_scale_x_from_point_if_not_initialized()
            key_exp_str: ExpressionString = scale_interface_helper.\
                get_coordinate_key_for_expression(coordinate=int(x._value))
            self._scale_x_from_point._value[key_exp_str.value] = scale_x
            self._append_scale_x_from_point_update_expression(x=x)

    def _append_scale_x_from_point_update_expression(
            self, *, x: Int) -> None:
        """
        Append the scale-x from the specified x-coordinate updating
        expression.

        Parameters
        ----------
        x : Int
            X-coordinate.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self._append_scale_x_from_point_update_expression,
                locals_=locals(),
                module_name=__name__, class_=ScaleXFromPointInterface):
            from apysc._display import scale_interface_helper
            expression: str
            expression = scale_interface_helper.get_scale_updating_expression(
                coordinate=x,
                scale_dict=self._scale_x_from_point,
                interface_variable_name=self.variable_name,
                coordinate_type=scale_interface_helper.CoordinateType.X)
            ap.append_js_expression(expression=expression)

    _scale_x_from_point_snapshots: Dict[str, Dict[str, Any]]

    def _make_snapshot(self, *, snapshot_name: str) -> None:
        """
        Make a value's snapshot.

        Parameters
        ----------
        snapshot_name : str
            Target snapshot name.
        """
        self._initialize_scale_x_from_point_if_not_initialized()
        self._set_single_snapshot_val_to_dict(
            dict_name='_scale_x_from_point_snapshots',
            value={**self._scale_x_from_point._value},
            snapshot_name=snapshot_name)

    def _revert(self, *, snapshot_name: str) -> None:
        """
        Revert a value if snapshot exists.

        Parameters
        ----------
        snapshot_name : str
            Target snapshot name.
        """
        if not self._snapshot_exists(snapshot_name=snapshot_name):
            return
        self._scale_x_from_point._value = self._scale_x_from_point_snapshots[
            snapshot_name]



"""Class implementation for the stop_propagation interface.
"""

from apysc._type.variable_name_interface import VariableNameInterface


class StopPropagationInterface(VariableNameInterface):

    def stop_propagation(self) -> None:
        """
        Stop event propagation.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self.stop_propagation, locals_=locals(),
                module_name=__name__, class_=StopPropagationInterface):
            expression: str = (
                f'{self.variable_name}.stopPropagation();'
            )
            ap.append_js_expression(expression=expression)



"""Class implementation for line alpha interface.
"""

from typing import Dict
from typing import Union

from apysc._animation.animation_line_alpha_interface import \
    AnimationLineAlphaInterface
from apysc._type.attr_linking_interface import AttrLinkingInterface
from apysc._type.number import Number
from apysc._type.revert_interface import RevertInterface


class LineAlphaInterface(
        AnimationLineAlphaInterface, RevertInterface, AttrLinkingInterface):

    _line_alpha: Number

    def _initialize_line_alpha_if_not_initialized(self) -> None:
        """
        Initialize _line_alpha attribute if it hasn't been initialized yet.
        """
        if hasattr(self, '_line_alpha'):
            return
        self._line_alpha = Number(1.0)

        self._append_line_alpha_attr_linking_setting()

    def _append_line_alpha_attr_linking_setting(self) -> None:
        """
        Append a line alpha attribute linking setting.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='line_alpha', locals_=locals(),
                module_name=__name__, class_=LineAlphaInterface):
            self._append_applying_new_attr_val_exp(
                new_attr=self._line_alpha, attr_name='line_alpha')
            self._append_attr_to_linking_stack(
                attr=self._line_alpha, attr_name='line_alpha')

    @property
    def line_alpha(self) -> Number:
        """
        Get this instance's line alpha (opacity).

        Returns
        -------
        line_alpha : Number
            Current line alpha (opacity. 0.0 to 1.0).

        References
        ----------
        - Graphics line_alpha interface document
            - https://simon-ritchie.github.io/apysc/graphics_line_alpha.html
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='line_alpha', locals_=locals(),
                module_name=__name__, class_=LineAlphaInterface):
            from apysc._type import value_util
            self._initialize_line_alpha_if_not_initialized()
            return value_util.get_copy(value=self._line_alpha)

    @line_alpha.setter
    def line_alpha(self, value: Number) -> None:
        """
        Update this instance's line alpha (opacity).

        Parameters
        ----------
        value : Number
            Line alpha (opacity) to set.

        References
        ----------
        - Graphics line_alpha interface document
            - https://simon-ritchie.github.io/apysc/graphics_line_alpha.html
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='line_alpha', locals_=locals(),
                module_name=__name__, class_=LineAlphaInterface):
            self._initialize_line_alpha_if_not_initialized()
            self._update_line_alpha_and_skip_appending_exp(value=value)
            self._line_alpha._append_incremental_calc_substitution_expression()
            self._append_line_alpha_update_expression()

            self._append_line_alpha_attr_linking_setting()

    def _append_line_alpha_update_expression(self) -> None:
        """
        Append line alpha updating expression.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self._append_line_alpha_update_expression,
                locals_=locals(),
                module_name=__name__, class_=LineAlphaInterface):
            from apysc._type import value_util
            line_alpha_str: str = value_util.get_value_str_for_expression(
                value=self._line_alpha)
            expression: str = (
                f'{self.variable_name}.stroke({{opacity: {line_alpha_str}}});'
            )
            ap.append_js_expression(expression=expression)

    def _update_line_alpha_and_skip_appending_exp(
            self, *, value: Union[float, Number]) -> None:
        """
        Update line alpha and skip appending expression.

        Parameters
        ----------
        value : float or Number
            Line alpha (opacity) to set.
        """
        from apysc._validation import color_validation
        from apysc._validation import number_validation
        number_validation.validate_num(num=value)
        color_validation.validate_alpha_range(alpha=value)
        if isinstance(value, float):
            value = Number(value)
        self._line_alpha = value

    _line_alpha_snapshots: Dict[str, float]

    def _make_snapshot(self, *, snapshot_name: str) -> None:
        """
        Make value's snapshot.

        Parameters
        ----------
        snapshot_name : str
            Target snapshot name.
        """
        self._initialize_line_alpha_if_not_initialized()
        self._set_single_snapshot_val_to_dict(
            dict_name='_line_alpha_snapshots',
            value=self._line_alpha._value, snapshot_name=snapshot_name)

    def _revert(self, *, snapshot_name: str) -> None:
        """
        Revert value if snapshot exists.

        Parameters
        ----------
        snapshot_name : str
            Target snapshot name.
        """
        if not self._snapshot_exists(snapshot_name=snapshot_name):
            return
        self._line_alpha._value = self._line_alpha_snapshots[snapshot_name]



"""Class implementation for the custom event interface.
"""

from typing import Any
from typing import Callable
from typing import Dict
from typing import Optional
from typing import Union

from apysc._event.custom_event_type import CustomEventType
from apysc._event.event import Event
from apysc._event.handler import HandlerData
from apysc._type.blank_object_interface import BlankObjectInterface

_CustomEventType = str
_HandlerName = str
_Handler = Callable[[Any, Any], None]


class CustomEventInterface(BlankObjectInterface):

    _custom_event_handlers: Dict[
        _CustomEventType,
        Dict[_HandlerName, HandlerData]]

    def _initialize_custom_event_handlers_if_not_initialized(
            self, *, custom_event_type_str: str) -> None:
        """
        Initialize the _custom_event_handlers data if it hasn't been
        initialized yet.

        Parameters
        ----------
        custom_event_type_str : str
            Target custom event type string.
        """
        if not hasattr(self, '_custom_event_handlers'):
            self._custom_event_handlers = {}
        if custom_event_type_str not in self._custom_event_handlers:
            self._custom_event_handlers[custom_event_type_str] = {}

    def _get_custom_event_type_str(
            self, *,
            custom_event_type: Union[CustomEventType, str]) -> str:
        """
        Get a custom event type string from a type value.

        Parameters
        ----------
        custom_event_type : CustomEventType or str
            Target custom event type or string.

        Returns
        -------
        custom_event_type_str : str
            A custom event type string.
        """
        if isinstance(custom_event_type, str):
            return custom_event_type
        custom_event_type_str: str = custom_event_type.value
        return custom_event_type_str

    def _set_custom_event_handler_data(
            self, *, handler: _Handler,
            custom_event_type_str: str,
            options: Optional[Any]) -> None:
        """
        Set a handler's data to the dictionary.

        Parameters
        ----------
        handler : _Handler
            Callable will be called when an event is dispatched.
        custom_event_type_str : str
            Target custom event type string.
        options : dict or None
            Optional arguments dictionary to be passed to a handler.
        """
        from apysc._event.handler import get_handler_name
        name: str = get_handler_name(handler=handler, instance=self)
        if options is None:
            options = {}
        self._custom_event_handlers[
            custom_event_type_str][name] = {  # type: ignore
                'handler': handler,
                'options': options,
        }

    def _unset_custom_event_handler_data(
            self, *, handler: _Handler,
            custom_event_type_str: str) -> None:
        """
        Unset a handler's data from the dictionary.

        Parameters
        ----------
        handler : _Handler
            Callable will be called when an event is dispatched.
        custom_event_type_str : str
            Target custom event type string.
        """
        from apysc._event.handler import get_handler_name
        if custom_event_type_str not in self._custom_event_handlers:
            return
        name: str = get_handler_name(handler=handler, instance=self)
        if name not in self._custom_event_handlers[custom_event_type_str]:
            return
        del self._custom_event_handlers[custom_event_type_str][name]

    def bind_custom_event(
            self, custom_event_type: Union[CustomEventType, str],
            handler: _Handler,
            e: Event,
            *,
            options: Optional[Any] = None,
            in_handler_head_expression: str = '') -> str:
        """
        Add a custom event listener setting.

        Parameters
        ----------
        custom_event_type : CustomEventType or str
            Target custom event type.
        handler : _Handler
            A handler will be called when the custom event is triggered.
        e : Event
            Event instance.
        options : dict or None, default None
            Optional arguments dictionary to be passed to a handler.
        in_handler_head_expression : str, default ''
            Optional expression to be added at the handler function's
            head position.

        Returns
        -------
        name : str
            Handler's name.

        References
        ----------
        - Bind and trigger the custom event document
            - https://bit.ly/3rky7VI
        - About the handler options’ type document
            - https://bit.ly/39tnYxC
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self.bind_custom_event, locals_=locals(),
                module_name=__name__, class_=CustomEventInterface):
            from apysc._event.handler import append_handler_expression
            from apysc._event.handler import get_handler_name
            custom_event_type_str: str = self._get_custom_event_type_str(
                custom_event_type=custom_event_type)
            self._initialize_custom_event_handlers_if_not_initialized(
                custom_event_type_str=custom_event_type_str)
            self._set_custom_event_handler_data(
                handler=handler, custom_event_type_str=custom_event_type_str,
                options=options)
            name: str = get_handler_name(handler=handler, instance=self)
            self._append_custom_event_binding_expression(
                custom_event_type_str=custom_event_type_str, name=name)
            handler_data: HandlerData = \
                self._custom_event_handlers[custom_event_type_str][name]
            append_handler_expression(
                handler_data=handler_data, handler_name=name, e=e,
                in_handler_head_expression=in_handler_head_expression)
            return name

    def _append_custom_event_binding_expression(
            self, *, custom_event_type_str: str, name: str) -> None:
        """
        Append a custom event binding expression.

        Parameters
        ----------
        custom_event_type_str : str
            Target custom event type string.
        name : str
            Handler's name.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self._append_custom_event_binding_expression,
                locals_=locals(),
                module_name=__name__, class_=CustomEventInterface):
            blank_object_variable_name: str = self.blank_object_variable_name
            expression: str = (
                f'$({blank_object_variable_name})'
                f'.off("{custom_event_type_str}", {name});'
                f'\n$({blank_object_variable_name})'
                f'.on("{custom_event_type_str}", {name});'
            )
            ap.append_js_expression(expression=expression)

    def trigger_custom_event(
            self, custom_event_type: Union[CustomEventType, str]) -> None:
        """
        Add a custom event trigger setting.

        Parameters
        ----------
        custom_event_type : CustomEventType or str
            Target custom event type.

        References
        ----------
        - Bind and trigger the custom event document
            - https://bit.ly/3rky7VI
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self.trigger_custom_event, locals_=locals(),
                module_name=__name__, class_=CustomEventInterface):
            blank_object_variable_name: str = self.blank_object_variable_name
            custom_event_type_str: str = self._get_custom_event_type_str(
                custom_event_type=custom_event_type)
            expression: str = (
                f'$({blank_object_variable_name})'
                f'.trigger("{custom_event_type_str}");'
            )
            ap.append_js_expression(expression=expression)

    def unbind_custom_event(
            self,
            custom_event_type: Union[CustomEventType, str],
            handler: _Handler) -> str:
        """
        Unbind (remove) a custom event listener setting.

        Parameters
        ----------
        custom_event_type : CustomEventType or str
            Target custom event type.
        handler : _Handler
            A handler will be called when the custom event is triggered.

        Returns
        -------
        name : str
            Handler's name.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self.unbind_custom_event, locals_=locals(),
                module_name=__name__, class_=CustomEventInterface):
            from apysc._event.handler import get_handler_name
            custom_event_type_str: str = self._get_custom_event_type_str(
                custom_event_type=custom_event_type)
            self._initialize_custom_event_handlers_if_not_initialized(
                custom_event_type_str=custom_event_type_str)
            self._unset_custom_event_handler_data(
                handler=handler, custom_event_type_str=custom_event_type_str)
            name: str = get_handler_name(handler=handler, instance=self)
            self._append_custom_event_unbinding_expression(
                custom_event_type_str=custom_event_type_str, name=name)
            return name

    def _append_custom_event_unbinding_expression(
            self, *, custom_event_type_str: str, name: str) -> None:
        """
        Add a custom event unbinding expression.

        Parameters
        ----------
        custom_event_type_str : str
            Target custom event type string.
        name : str
            Handler's name.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self._append_custom_event_unbinding_expression,
                locals_=locals(),
                module_name=__name__, class_=CustomEventInterface):
            expression: str = (
                f'$({self.blank_object_variable_name})'
                f'.off("{custom_event_type_str}", {name});'
            )
            ap.append_js_expression(expression=expression)

    def unbind_custom_event_all(
            self, custom_event_type: Union[CustomEventType, str]) -> None:
        """
        Unbind (remove) custom event listener settings.

        Parameters
        ----------
        custom_event_type : CustomEventType or str
            Target custom event type.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self.unbind_custom_event_all, locals_=locals(),
                module_name=__name__, class_=CustomEventInterface):
            custom_event_type_str: str = self._get_custom_event_type_str(
                custom_event_type=custom_event_type)
            self._initialize_custom_event_handlers_if_not_initialized(
                custom_event_type_str=custom_event_type_str)
            self._custom_event_handlers[custom_event_type_str] = {}
            expression: str = (
                f'$({self.blank_object_variable_name})'
                f'.off("{custom_event_type_str}");'
            )
            ap.append_js_expression(expression=expression)



"""
The module that is implemented each Jupyter interface and definition.

Mainly the following interfaces are defined:

- display_on_jupyter
    Save the overall HTML and display it on the Jupyter.
- display_on_colaboratory
    Save the overall HTML and display it on the Google Colaboratory.
"""

import os
import shutil
from datetime import datetime
from random import randint

_TMP_ROOT_DIR_PATH: str = './'


def _save_overall_html(*, html_file_name: str, minify: bool) -> None:
    """
    Save the overall HTML file.

    Parameters
    ----------
    html_file_name : str, default 'index.html'
        The output HTML file name.
    minify : bool, default True
        Boolean value whether minify a HTML or not.
    """
    import apysc as ap
    timestamp: float = datetime.now().timestamp()
    random_int: int = randint(10000, 100000)
    dest_dir_path: str = (
        f'{_TMP_ROOT_DIR_PATH}tmp_apysc_jupyter_{timestamp}{random_int}/')
    ap.save_overall_html(
        dest_dir_path=dest_dir_path,
        html_file_name=html_file_name,
        minify=minify,
        skip_js_lib_exporting=True,
        embed_js_libs=True,
        verbose=0)
    src_file_path: str = os.path.join(dest_dir_path, html_file_name)
    shutil.copy(src_file_path, html_file_name)
    shutil.rmtree(dest_dir_path, ignore_errors=True)


def display_on_jupyter(
        html_file_name: str,
        *,
        minify: bool = True) -> None:
    """
    Save the overall HTML and display it on the Jupyter.

    Notes
    -----
    Currently not supported the Jupyter on the VS Code.

    Parameters
    ----------
    html_file_name : str, default 'index.html'
        The output HTML file name.
    minify : bool, default True
        Boolean value whether minify a HTML or not.
        False setting is useful when debugging.

    Notes
    -----
    This interface requires the Jupyter library (e.g., `notebook` package).

    References
    ----------
    - display_on_jupyter interface document
        - https://simon-ritchie.github.io/apysc/display_on_jupyter.html
    """
    from IPython.display import IFrame
    from IPython.display import display

    import apysc as ap

    stage: ap.Stage = ap.get_stage()
    _save_overall_html(
        html_file_name=html_file_name,
        minify=minify)
    display(
        IFrame(src=f'./{html_file_name}',
               width=stage._width._value, height=stage._height._value))


def display_on_colaboratory(
        html_file_name: str, *,
        minify: bool = True) -> None:
    """
    Save the overall HTML and display it on the Google Colaboratory.

    Parameters
    ----------
    html_file_name : str, default 'index.html'
        The output HTML file name.
    minify : bool, default True
        Boolean value whether minify a HTML or not.
        False setting is useful when debugging.

    References
    ----------
    display_on_colaboratory interface document
        https://simon-ritchie.github.io/apysc/display_on_colaboratory.html
    """
    from IPython.display import HTML
    from IPython.display import display

    from apysc._file import file_util
    _save_overall_html(
        html_file_name=html_file_name,
        minify=minify)
    html_str: str = file_util.read_txt(file_path=html_file_name)
    display(HTML(html_str))



"""Class implementation for the y-coordinate interface.
"""

from typing import Dict

from apysc._animation.animation_move_interface import AnimationMoveInterface
from apysc._animation.animation_y_interface import AnimationYInterface
from apysc._type.attr_linking_interface import AttrLinkingInterface
from apysc._type.int import Int
from apysc._type.revert_interface import RevertInterface


class YInterface(
        AnimationYInterface, AnimationMoveInterface, RevertInterface,
        AttrLinkingInterface):

    _y: Int

    def _initialize_y_if_not_initialized(self) -> None:
        """
        Initialize the _y attribute if it hasn't been initialized yet.
        """
        if hasattr(self, '_y'):
            return
        self._y = Int(0)

        self._append_y_attr_linking_setting()

    def _append_y_attr_linking_setting(self) -> None:
        """
        Append a y attribute linking setting.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self._append_y_attr_linking_setting,
                locals_=locals(),
                module_name=__name__, class_=YInterface):
            self._append_applying_new_attr_val_exp(
                new_attr=self._y, attr_name='y')
            self._append_attr_to_linking_stack(
                attr=self._y, attr_name='y')

    @property
    def y(self) -> Int:
        """
        Get a y-coordinate.

        Returns
        -------
        y : Int
            Y-coordinate.

        References
        ----------
        - Display object x and y interfaces document
            - https://bit.ly/2ToA5ba
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='y', locals_=locals(),
                module_name=__name__, class_=YInterface):
            from apysc._type import value_util
            self._initialize_y_if_not_initialized()
            y: ap.Int = value_util.get_copy(value=self._y)
            return y

    @y.setter
    def y(self, value: Int) -> None:
        """
        Update y-coordinate.

        Parameters
        ----------
        value : int or Int
            Y-coordinate value.

        References
        ----------
        - Display object x and y interfaces document
            - https://bit.ly/2ToA5ba
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='y', locals_=locals(),
                module_name=__name__, class_=YInterface):
            from apysc._type.number_value_interface import NumberValueInterface
            from apysc._validation import number_validation
            if not isinstance(value, NumberValueInterface):
                number_validation.validate_integer(integer=value)
                value = ap.Int(value=value)
            self._y = value
            self._y._append_incremental_calc_substitution_expression()
            self._append_y_update_expression()

            self._append_y_attr_linking_setting()

    def _append_y_update_expression(self) -> None:
        """
        Append y position updating expression.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self._append_y_update_expression, locals_=locals(),
                module_name=__name__, class_=YInterface):
            from apysc._type import value_util
            self._initialize_y_if_not_initialized()
            value_str: str = value_util.get_value_str_for_expression(
                value=self._y)
            expression: str = (
                f'{self.variable_name}.y({value_str});'
            )
            ap.append_js_expression(expression=expression)

    _y_snapshots: Dict[str, int]

    def _make_snapshot(self, *, snapshot_name: str) -> None:
        """
        Make a value's snapshot.

        Parameters
        ----------
        snapshot_name : str
            Target snapshot name.
        """
        self._initialize_y_if_not_initialized()
        self._set_single_snapshot_val_to_dict(
            dict_name='_y_snapshots',
            value=int(self._y._value), snapshot_name=snapshot_name)

    def _revert(self, *, snapshot_name: str) -> None:
        """
        Revert a value if snapshot exists.

        Parameters
        ----------
        snapshot_name : str
            Target snapshot name.
        """
        if not self._snapshot_exists(snapshot_name=snapshot_name):
            return
        self._y._value = self._y_snapshots[snapshot_name]



"""The common utility implementations for the path data.
"""

from typing import List

from apysc._geom.path_data_base import PathDataBase


def make_paths_expression_from_list(
        *, path_data_list: List[PathDataBase]) -> str:
    """
    Make a paths expression from a specified list of path data.

    Parameters
    ----------
    path_data_list : list of PathDataBase
        Target path data settings, such as the ap.PathData.MoveTo.

    Returns
    -------
    expression : str
        Created expression string.
    """
    import apysc as ap
    with ap.DebugInfo(
            callable_=make_paths_expression_from_list,
            locals_=locals(), module_name=__name__):
        expression: str = ''
        for path_data in path_data_list:
            if expression != '':
                expression += ' + " " + '
            expression += path_data._get_svg_str()
        return expression



"""Class implementation for line round dot setting interface.
"""

from typing import Dict
from typing import Optional

from apysc._display.line_cap_interface import LineCapInterface
from apysc._display.line_round_dot_setting import LineRoundDotSetting
from apysc._display.line_thickness_interface import LineThicknessInterface


class LineRoundDotSettingInterface(LineCapInterface, LineThicknessInterface):

    _line_round_dot_setting: Optional[LineRoundDotSetting]

    def _initialize_line_round_dot_setting_if_not_initialized(self) -> None:
        """
        Initialize _line_round_dot_setting if it is not
        initialized yet.
        """
        if hasattr(self, '_line_round_dot_setting'):
            return
        self._line_round_dot_setting = None

    @property
    def line_round_dot_setting(self) -> Optional[LineRoundDotSetting]:
        """
        Get this instance's line round dot setting.

        Returns
        -------
        line_round_dot_setting : LineRoundDotSetting or None
            Line round dot setting.

        References
        ----------
        - Graphics line_round_dot interface document
            - https://bit.ly/3Deklci
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='line_round_dot_setting', locals_=locals(),
                module_name=__name__, class_=LineRoundDotSettingInterface):
            self._initialize_line_round_dot_setting_if_not_initialized()
            return self._line_round_dot_setting

    @line_round_dot_setting.setter
    def line_round_dot_setting(
            self, value: Optional[LineRoundDotSetting]) -> None:
        """
        Set line round dot setting.

        Notes
        -----
        This property updating will affect line cap and
        thickness settings.

        Parameters
        ----------
        value : LineRoundDotSetting or None
            Line round setting to set.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='line_round_dot_setting', locals_=locals(),
                module_name=__name__, class_=LineRoundDotSettingInterface):
            from apysc._validation import display_validation
            self._update_line_round_dot_setting_and_skip_appending_exp(
                value=value)
            if value is not None:
                self.line_cap = ap.LineCaps.ROUND
                self.line_thickness = value.round_size
            else:
                self.line_cap = ap.LineCaps.BUTT
            self._append_line_round_dot_setting_update_expression()
            display_validation.validate_multiple_line_settings_isnt_set(
                any_instance=self)

    def _update_line_round_dot_setting_and_skip_appending_exp(
            self, *, value: Optional[LineRoundDotSetting]) -> None:
        """
        Update line round setting and skip appending expression.

        Parameters
        ----------
        value : LineRoundSetting or None
            Line round dot setting to set.
        """
        if value is not None and not isinstance(value, LineRoundDotSetting):
            raise TypeError(
                'Not supported line_round_dot_setting type specified: '
                f'{type(value)}'
                '\nAcceptable ones are: LineRoundDotSetting or None.')
        self._line_round_dot_setting = value

    def _append_line_round_dot_setting_update_expression(self) -> None:
        """
        Append line round dot setting updating expression.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self._append_line_round_dot_setting_update_expression,  # noqa
                locals_=locals(),
                module_name=__name__, class_=LineRoundDotSettingInterface):
            import apysc as ap
            if self._line_round_dot_setting is None:
                setting_str: str = '""'
            else:
                round_size_name: str = \
                    self._line_round_dot_setting.round_size.variable_name
                space_size_name: str = \
                    self._line_round_dot_setting.space_size.variable_name
                setting_str = (
                    f'"1 " + String({round_size_name} + {space_size_name})'
                )
            expression: str = (
                f'{self.variable_name}.css("stroke-dasharray", {setting_str});'
            )
            ap.append_js_expression(expression=expression)

    _line_round_dot_setting_snapshots: Dict[
        str, Optional[LineRoundDotSetting]]

    def _make_snapshot(self, *, snapshot_name: str) -> None:
        """
        Make value's snapshot.

        Parameters
        ----------
        snapshot_name : str
            Target snapshot name.
        """
        self._initialize_line_round_dot_setting_if_not_initialized()
        self._set_single_snapshot_val_to_dict(
            dict_name='_line_round_dot_setting_snapshots',
            value=self._line_round_dot_setting, snapshot_name=snapshot_name)

    def _revert(self, *, snapshot_name: str) -> None:
        """
        Revert value if snapshot exists.

        Parameters
        ----------
        snapshot_name : str
            Target snapshot name.
        """
        if not self._snapshot_exists(snapshot_name=snapshot_name):
            return
        self._line_round_dot_setting = self._line_round_dot_setting_snapshots[
            snapshot_name]



"""Class implementation for the css interface.
"""

from typing import Dict
from typing import Union

from apysc._type.revert_interface import RevertInterface
from apysc._type.string import String
from apysc._type.variable_name_interface import VariableNameInterface


class CssInterface(VariableNameInterface, RevertInterface):

    _css: Dict[str, String]

    def _initialize_css_if_not_initialized(self) -> None:
        """
        Initialize the _css attribute if it hasn't been initialized yet.
        """
        if hasattr(self, '_css'):
            return
        self._css = {}

    def get_css(self, name: Union[str, String]) -> String:
        """
        Get a CSS value string.

        Parameters
        ----------
        name : str or String
            CSS name (e.g., 'display').

        Returns
        -------
        css : ap.String
            CSS value.

        References
        ----------
        Display object get_css and set_css interfaces document
            https://bit.ly/3wNgIWI

        Examples
        --------
        >>> import apysc as ap
        >>> stage: ap.Stage = ap.Stage()
        >>> sprite: ap.Sprite = ap.Sprite()
        >>> sprite.graphics.begin_fill(color='#0af', alpha=0.5)
        >>> sprite.set_css(name='display', value='none')
        >>> sprite.get_css(name='display')
        String('none')
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self.get_css, locals_=locals(),
                module_name=__name__, class_=CssInterface):
            from apysc._converter import to_builtin_val_from_apysc
            self._initialize_css_if_not_initialized()
            name_: str = to_builtin_val_from_apysc.\
                get_builtin_str_from_apysc_val(string=name)
            if name_ in self._css:
                css: ap.String = self._css[name_]._copy()
            else:
                css = ap.String('')
            self._append_get_css_expresion(name=name, css=css)
            return css

    def _append_get_css_expresion(
            self, *, name: Union[str, String], css: String) -> None:
        """
        Append a css getter expression string.

        Parameters
        ----------
        name : str or String
            CSS name (e.g., 'display').
        css : String
            CSS value.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self._append_get_css_expresion, locals_=locals(),
                module_name=__name__, class_=CssInterface):
            from apysc._type import value_util
            name_value_str: str = value_util.get_value_str_for_expression(
                value=name)
            css_value_str: str = value_util.get_value_str_for_expression(
                value=css)
            expression: str = (
                f'{css_value_str} = {self.variable_name}.'
                f'css({name_value_str});'
            )
            ap.append_js_expression(expression=expression)

    def set_css(
            self, name: Union[str, String],
            value: Union[str, String]) -> None:
        """
        Set a specified value string to the CSS.

        Parameters
        ----------
        name : str or String
            CSS name (e.g., 'display').
        value : str or String
            A CSS value string (e.g., 'none').

        References
        ----------
        Display object get_css and set_css interfaces document
            https://bit.ly/3wNgIWI

        Examples
        --------
        >>> import apysc as ap
        >>> stage: ap.Stage = ap.Stage()
        >>> sprite: ap.Sprite = ap.Sprite()
        >>> sprite.graphics.begin_fill(color='#0af', alpha=0.5)
        >>> sprite.set_css(name='display', value='none')
        >>> sprite.get_css(name='display')
        String('none')
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self.set_css, locals_=locals(),
                module_name=__name__, class_=CssInterface):
            from apysc._converter import to_apysc_val_from_builtin
            from apysc._converter import to_builtin_val_from_apysc
            self._initialize_css_if_not_initialized()
            name_: str = to_builtin_val_from_apysc.\
                get_builtin_str_from_apysc_val(string=name)
            value_: ap.String = to_apysc_val_from_builtin.\
                get_copied_string_from_builtin_val(string=value)
            self._css[name_] = value_
            self._append_set_css_expression(name=name, value=value)

    def _append_set_css_expression(
            self, *, name: Union[str, String],
            value: Union[str, String]) -> None:
        """
        Append a css setter expression string.

        Parameters
        ----------
        name : str or String
            CSS name (e.g., 'display').
        value : str or String
            A CSS value string (e.g., 'none').
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self._append_set_css_expression, locals_=locals(),
                module_name=__name__, class_=CssInterface):
            from apysc._type import value_util
            name_value_str: str = value_util.get_value_str_for_expression(
                value=name)
            value_str: str = value_util.get_value_str_for_expression(
                value=value)
            expression: str = (
                f'{self.variable_name}.css({name_value_str}, {value_str});'
            )
            ap.append_js_expression(expression=expression)

    _css_snapshot: Dict[str, Dict[str, String]]

    def _make_snapshot(self, *, snapshot_name: str) -> None:
        """
        Make values' snapshot.

        Parameters
        ----------
        snapshot_name : str
            Target snapshot name.
        """
        self._initialize_css_if_not_initialized()
        self._set_single_snapshot_val_to_dict(
            dict_name='_css_snapshot',
            value={**self._css}, snapshot_name=snapshot_name)

    def _revert(self, *, snapshot_name: str) -> None:
        """
        Revert values if snapshot exists.

        Parameters
        ----------
        snapshot_name : str
            Target snapshot name.
        """
        if not self._snapshot_exists(snapshot_name=snapshot_name):
            return
        self._css = self._css_snapshot[snapshot_name]



"""Class implementation for parent related interface.
"""

from typing import Any
from typing import Dict
from typing import Optional

from apysc._type.revert_interface import RevertInterface


class ParentInterface(RevertInterface):

    _parent: Optional[Any] = None

    @property
    def parent(self) -> Optional[Any]:
        """
        Get parent instance that has a add_child and remove_child
        interfaces.

        Returns
        -------
        parent : any parent instance or None
            Parent instance that has a add_child and remove_child
            interfaces. If this instance not have parent instance (not
            added child), None will be returned.

        References
        ----------
        - Display object parent interfaces document
            - https://bit.ly/3wQX782
        """
        return self._parent

    @parent.setter
    def parent(self, value: Optional[Any]) -> None:
        """
        Set parent instance.

        Notes
        -----
        Only use this setter method at add_child or that's related
        interface.

        Parameters
        ----------
        value : *
            Parent instance to be set.

        Raises
        ------
        ValueError
            If specified instance is not None and hasn't `ChildInterface`
            interfaces.

        References
        ----------
        - Display object parent interfaces document
            - https://bit.ly/3wQX782
        """
        from apysc._validation import parent_validation
        parent_validation.validate_parent_instance(parent=value)
        parent_validation.validate_parent_contains_child(
            parent=value, child=self)
        self._parent = value

    def remove_from_parent(self) -> None:
        """
        Remove this instance from parent.

        Raises
        ------
        ValueError
            If this instance is not added to any parent.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self.remove_from_parent, locals_=locals(),
                module_name=__name__, class_=ParentInterface):
            from apysc._display.child_interface import ChildInterface
            from apysc._display.child_interface import \
                append_expression_of_remove_child
            from apysc._display.display_object import DisplayObject
            parent: Optional[ChildInterface] = self._parent
            child: DisplayObject = self  # type: ignore
            if parent is not None:
                parent.remove_child(child=child)
            else:
                append_expression_of_remove_child(child=child)

    _parent_snapshots: Dict[str, Optional[Any]]

    def _make_snapshot(self, *, snapshot_name: str) -> None:
        """
        Make value's snapshot.

        Parameters
        ----------
        snapshot_name : str
            Target snapshot name.
        """
        self._set_single_snapshot_val_to_dict(
            dict_name='_parent_snapshots',
            value=self._parent, snapshot_name=snapshot_name)

    def _revert(self, *, snapshot_name: str) -> None:
        """
        Revert value if snapshot exists.

        Parameters
        ----------
        snapshot_name : str
            Target snapshot name.
        """
        if not self._snapshot_exists(snapshot_name=snapshot_name):
            return
        self._parent = self._parent_snapshots[snapshot_name]



"""Helper module for the scale interfaces.
"""

from enum import Enum
from typing import Union

from apysc._type.dictionary import Dictionary
from apysc._type.expression_string import ExpressionString
from apysc._type.int import Int
from apysc._type.number import Number


def get_coordinate_key_for_expression(
        *,
        coordinate: Union[int, Int]) -> ExpressionString:
    """
    Get a key string for the expression from the x or y coordinate.

    Parameters
    ----------
    coordinate : int or Int
        X or y coordinate.

    Returns
    -------
    key_exp_str : ExpressionString
        Key expression string.
    """
    from apysc._type.variable_name_interface import VariableNameInterface
    if isinstance(coordinate, VariableNameInterface):
        coordinate_str: str = coordinate.variable_name
    else:
        coordinate_str = str(coordinate)
    key_exp_str: ExpressionString = ExpressionString(
        value=f'String({coordinate_str})')
    return key_exp_str


class CoordinateType(Enum):
    X = 1
    Y = 2


def get_scale_updating_expression(
        *,
        coordinate: Int,
        scale_dict: Dictionary[str, Number],
        interface_variable_name: str,
        coordinate_type: CoordinateType) -> str:
    """
    Get a scale updating expression string from a specified coordinate.

    Parameters
    ----------
    coordinate : Int
        X or y coordinate.
    scale_dict : Dictionary
        Scale value dictionary.
    interface_variable_name : str
        Scale interface instance variable name.
    coordinate_type : CoordinateType
        Coordinate type to identify the x or y target.

    Returns
    -------
    expression : str
        A scale updating expression string.
    """
    from apysc._display import scale_interface_helper
    from apysc._expression import expression_variables_util
    from apysc._expression import var_names
    from apysc._type import value_util
    before_value_name: str = expression_variables_util.\
        get_next_variable_name(type_name=var_names.NUMBER)
    key_exp_str_1: ExpressionString = scale_interface_helper.\
        get_coordinate_key_for_expression(coordinate=coordinate)
    key_exp_str_2: ExpressionString = scale_interface_helper.\
        get_coordinate_key_for_expression(coordinate=int(coordinate._value))
    after_value_str: str = value_util.get_value_str_for_expression(
        value=scale_dict._value[key_exp_str_2.value])
    coordinate_value_str: str = value_util.get_value_str_for_expression(
        value=coordinate)
    scale_from_point_value_str: str = value_util.\
        get_value_str_for_expression(value=scale_dict)
    if coordinate_type == CoordinateType.X:
        coordinate_exp: str = f'{coordinate_value_str}, 0'
        scale_resetting_exp: str = f'1 / {before_value_name}, 1'
        scale_result_exp: str = f'{after_value_str}, 1'
    else:
        coordinate_exp = f'0, {coordinate_value_str}'
        scale_resetting_exp = f'1, 1 / {before_value_name}'
        scale_result_exp = f'1, {after_value_str}'
    expression: str = (
        f'if ({key_exp_str_1.value} in '
        f'{scale_from_point_value_str}) {{'
        f'\n  var {before_value_name} = '
        f'{scale_from_point_value_str}[{key_exp_str_1.value}];'
        '\n}else {'
        f'\n  {before_value_name} = 1.0;'
        '\n}'
        f'\n{interface_variable_name}.scale('
        f'{scale_resetting_exp}, {coordinate_exp});'
        f'\n{interface_variable_name}.scale('
        f'{scale_result_exp}, {coordinate_exp});'
        f'\n{scale_from_point_value_str}[{key_exp_str_1.value}] = '
        f'{after_value_str};'
    )
    return expression



"""Class implementation for the animation_reverse interface.
"""

from apysc._type.variable_name_interface import VariableNameInterface


class AnimationReverseInterface(VariableNameInterface):

    def animation_reverse(self) -> None:
        """
        Reverse the all running animations.

        Notes
        -----
        If you call this interface multiple times and animations
        have been reached at the beginning or end of the animation,
        then the reverse instruction will be ignored.
        This means that the same interval's timer tick reverse
        setting will not work correctly (since the same interval
        setting will reach the animation start).

        References
        ----------
        - animation_reverse interface document
            - https://simon-ritchie.github.io/apysc/animation_reverse.html
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self.animation_reverse, locals_=locals(),
                module_name=__name__, class_=AnimationReverseInterface):
            expression: str = (
                f'{self.variable_name}.timeline().reverse();'
            )
            ap.append_js_expression(expression=expression)



"""Path data class implementation for the svg's `horizontal line` (H).
"""

from typing import Union

from apysc._geom.path_data_base import PathDataBase
from apysc._geom.path_x_interface import PathXInterface
from apysc._type.boolean import Boolean
from apysc._type.int import Int
from apysc._type.string import String


class PathHorizontal(PathDataBase, PathXInterface):
    """
    Path data class for the svg's `horizontal line` (H).
    """

    def __init__(
            self, x: Union[int, Int], *,
            relative: Union[bool, Boolean] = False) -> None:
        """
        Path data class for the svg's `horizontal line` (H).

        Parameters
        ----------
        x : int or Int
            X-coordinate of the destination point.
        relative : bool or Boolean, default False
            The boolean value indicating whether the path
            coordinates are relative or not (absolute).
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='__init__', locals_=locals(),
                module_name=__name__, class_=PathHorizontal):
            from apysc._converter.to_apysc_val_from_builtin import \
                get_copied_int_from_builtin_val
            from apysc._geom.path_label import PathLabel
            super(PathHorizontal, self).__init__(
                path_label=PathLabel.HORIZONTAL,
                relative=relative)
            self.x = get_copied_int_from_builtin_val(integer=x)

    def _get_svg_str(self) -> str:
        """
        Get a path's SVG string created with the current setting.

        Returns
        -------
        svg_str : str
            A path's SVG string created with the current setting.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self._get_svg_str, locals_=locals(),
                module_name=__name__, class_=PathHorizontal):
            from apysc._type import value_util
            svg_char: String = self._get_svg_char()
            svg_char_str: str = value_util.get_value_str_for_expression(
                value=svg_char)
            x_str: str = value_util.get_value_str_for_expression(
                value=self._x)
            svg_str: str = f'{svg_char_str} + String({x_str})'
            return svg_str

    def update_path_data(
            self, x: Union[int, Int],
            relative: Union[bool, Boolean]) -> None:
        """
        Update the path's data settings.

        Parameters
        ----------
        x : int or Int
            X-coordinate of the destination point.
        relative : bool or Boolean, default False
            The boolean value indicating whether the path
            coordinates are relative or not (absolute).
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self.update_path_data, locals_=locals(),
                module_name=__name__, class_=PathHorizontal):
            from apysc._converter.to_apysc_val_from_builtin import \
                get_copied_boolean_from_builtin_val
            from apysc._converter.to_apysc_val_from_builtin import \
                get_copied_int_from_builtin_val
            self.x = get_copied_int_from_builtin_val(integer=x)
            self.relative = get_copied_boolean_from_builtin_val(
                bool_val=relative)



"""Class implementation for handler.
"""

from typing import Any
from typing import Callable
from typing import List

from typing_extensions import TypedDict

from apysc._event.event import Event
from apysc._event.mouse_event_type import MouseEventType
from apysc._type.variable_name_interface import VariableNameInterface

_Handler = Callable[[Any, Any], None]


class HandlerData(TypedDict):
    handler: Callable[[Event, Any], None]
    options: Any


def get_handler_name(
        *, handler: _Handler, instance: Any) -> str:
    """
    Get a handler name.

    Parameters
    ----------
    handler : _Handler
        Target handler.
    instance : VariableNameInterface
        Instance to bind target handler.

    Returns
    -------
    handler_name : str
        Handler name (method path + class name (if handler is method)
        + function or method name) + instance's variable name.
    """
    from apysc._callable import callable_util
    from apysc._event import handler_circular_calling_util as circ_util
    from apysc._validation.variable_name_validation import \
        validate_variable_name_interface_type
    class_name: str = callable_util.get_method_class_name(method=handler)
    if class_name != '':
        class_name = f'{class_name}_'
    handler_name: str = (
        f'{handler.__module__}{class_name}{handler.__name__}'  # type: ignore
    )
    handler_name = handler_name.replace('.', '_')
    instance_: VariableNameInterface = validate_variable_name_interface_type(
        instance=instance)
    handler_name += f'_{instance_.variable_name}'
    if circ_util.is_handler_circular_calling(handler_name=handler_name):
        return circ_util.get_prev_handler_name(handler_name=handler_name)
    return handler_name


def append_handler_expression(
        *,
        handler_data: HandlerData,
        handler_name: str,
        e: Event,
        in_handler_head_expression: str = '') -> None:
    """
    Append a handler's expression.

    Parameters
    ----------
    handler_data : HandlerData
        Target handler's data to append.
    handler_name : str
        Target handler's name.
    e : Event
        Created event instance.
    in_handler_head_expression : str, default ''
        Optional expression to be added at the handler function's
        head position.
    """
    import apysc as ap
    with ap.DebugInfo(
            callable_=append_handler_expression, locals_=locals(),
            module_name=__name__):
        from apysc._event.handler_circular_calling_util import \
            is_handler_circular_calling
        from apysc._expression.event_handler_scope import HandlerScope
        from apysc._expression.indent_num import Indent
        from apysc._type import revert_interface
        from apysc._validation.event_validation import validate_event
        from apysc._validation.variable_name_validation import \
            validate_variable_name_interface_type
        validate_event(e=e)
        variables: List[Any] = [*handler_data['options'].values()]
        snapshot_name: str = revert_interface.make_variables_snapshots(
            variables=variables)
        instance: VariableNameInterface = \
            validate_variable_name_interface_type(instance=e.this)

        with HandlerScope(handler_name=handler_name, instance=instance):
            is_handler_circular_calling_: bool = is_handler_circular_calling(
                handler_name=handler_name)
            if not is_handler_circular_calling_:
                expression: str = (
                    f'function {handler_name}({e.variable_name}) {{'
                )
                ap.append_js_expression(expression=expression)
                with Indent():
                    _append_in_handler_head_expression(
                        in_handler_head_expression=in_handler_head_expression)
                    handler_data['handler'](
                        e, handler_data['options'])
                ap.append_js_expression(expression='}')

        revert_interface.revert_variables(
            snapshot_name=snapshot_name, variables=variables)


def _append_in_handler_head_expression(
        *, in_handler_head_expression: str) -> None:
    """
    Append an in-handler head expression if it is not blank.

    Parameters
    ----------
    in_handler_head_expression : str
        Optional expression to be added at the handler function's
        head position if it is not blank.
    """
    import apysc as ap
    with ap.DebugInfo(
            callable_=_append_in_handler_head_expression, locals_=locals(),
            module_name=__name__):
        if in_handler_head_expression == '':
            return
        ap.append_js_expression(expression=in_handler_head_expression)


def append_unbinding_expression(
        *, this: VariableNameInterface, handler_name: str,
        mouse_event_type: MouseEventType) -> None:
    """
    Append event unbinding expression.

    Parameters
    ----------
    this : VariableNameInterface
        Instance that event is binded.
    handler_name : str
        Target handler's name.
    mouse_event_type : MouseEventType
        Event type to unbind.
    """
    import apysc as ap
    with ap.DebugInfo(
            callable_=append_unbinding_expression, locals_=locals(),
            module_name=__name__):
        from apysc._validation import event_validation
        event_validation.validate_event_type(
            mouse_event_type=mouse_event_type)
        expression: str = (
            f'{this.variable_name}.off("{mouse_event_type.value}", '
            f'{handler_name});'
        )
        ap.append_js_expression(expression=expression)


def append_unbinding_all_expression(
        *, this: VariableNameInterface,
        mouse_event_type: MouseEventType) -> None:
    """
    Append all events unbinding expression.

    Parameters
    ----------
    this : VariableNameInterface
        Instance that events are binded.
    mouse_event_type : MouseEventType
        Event type to unbind.
    """
    import apysc as ap
    with ap.DebugInfo(
            callable_=append_unbinding_all_expression, locals_=locals(),
            module_name=__name__):
        from apysc._validation import event_validation
        event_validation.validate_event_type(
            mouse_event_type=mouse_event_type)
        expression: str = (
            f'{this.variable_name}.off("{mouse_event_type.value}");'
        )
        ap.append_js_expression(expression=expression)



"""Debugging mode setting interface implementations for the HTML
and JavaScript.
"""

import inspect
from typing import Any
from typing import Callable
from typing import Dict
from typing import Optional
from typing import Tuple
from typing import Type
from typing import Union

from apysc._expression.indent_num import Indent


def set_debug_mode() -> None:
    """
    Set the debug mode for the HTML and JavaScript debugging.
    If this functions is called, the following setting will be applied:
    - HTML minify setting will be disabled.
    - Per each interface JavaScript divider string will be appended.
    """
    from apysc._expression import expression_data_util
    table_name: str = expression_data_util.TableName.DEBUG_MODE_SETTING.value
    query: str = f'DELETE FROM {table_name};'
    expression_data_util.exec_query(sql=query, commit=False)
    query = f'INSERT INTO {table_name}(is_debug_mode) VALUES(1);'
    expression_data_util.exec_query(sql=query)


def unset_debug_mode() -> None:
    """
    Unset the debug mode for the HTML and JavaScript debugging.
    """
    from apysc._expression import expression_data_util
    table_name: str = expression_data_util.TableName.DEBUG_MODE_SETTING.value
    query: str = f'DELETE FROM {table_name};'
    expression_data_util.exec_query(sql=query)


def is_debug_mode() -> bool:
    """
    Get a boolean value whether the current debug mode is enabled or not.

    Returns
    -------
    result : bool
        If the current debug mode is enabled, True will be returned.
    """
    from apysc._expression import expression_data_util
    table_name: str = expression_data_util.TableName.DEBUG_MODE_SETTING.value
    query: str = f'SELECT is_debug_mode FROM {table_name} LIMIT 1;'
    expression_data_util.exec_query(sql=query)
    result_: Optional[Tuple[int]] = expression_data_util.cursor.fetchone()
    if result_ is None:
        return False
    result: bool = bool(result_[0])
    return result


def _get_callable_str(*, callable_: Union[Callable, str]) -> str:
    """
    Get a callable string (label).

    Parameters
    ----------
    callable_ : Callable or str
        Target function or method or property or dunder method name.

    Returns
    -------
    callable_str : str
        A callable string (label).
    """
    if isinstance(callable_, str):
        callable_str: str = callable_
    else:
        callable_str = callable_.__name__
    return callable_str


def _get_callable_path_name(
        *,
        callable_: Union[Callable, str],
        module_name: str,
        class_: Optional[Type] = None) -> str:
    """
    Get a specified callable count data path name.

    Parameters
    ----------
    callable_ : Callable or str
        Target function or method or property or dunder method name.
    module_name : str
        Module name. This value will be set the `__name__` value.
    class_ : Type or None, optional
        Target class type. If the target callable_ variable is not
        a method, this argument will be ignored.

    Returns
    -------
    path_name : str
        Target path name.
    """
    module_path: str = module_name.replace('.', '_')
    if class_ is None:
        class_name: str = ''
    else:
        class_name = f'_{class_.__name__}'
    callable_str: str = _get_callable_str(callable_=callable_)
    path_name: str = f'{module_path}{class_name}_{callable_str}'
    return path_name


def _get_callable_count(
        *,
        callable_: Union[Callable, str],
        module_name: str,
        class_: Optional[Type] = None) -> int:
    """
    Get a specified callable count number.

    Parameters
    ----------
    callable_ : Callable or str
        Target function or method or property or dunder method name.
    module_name : str
        Module name. This value will be set the `__name__` value.
    class_ : Type or None, optional
        Target class type. If the target callable_ variable is not
        a method, this argument will be ignored.

    Returns
    -------
    callable_count : int
        Target count number.
    """
    from apysc._expression import expression_data_util
    path_name: str = _get_callable_path_name(
        callable_=callable_, module_name=module_name, class_=class_)
    table_name: str = \
        expression_data_util.TableName.DEBUG_MODE_CALLABLE_COUNT.value
    query: str = (
        f'SELECT count FROM {table_name} '
        f"WHERE name = '{path_name}' LIMIT 1;"
    )
    expression_data_util.exec_query(sql=query)
    result: Optional[Tuple[int]] = expression_data_util.cursor.fetchone()
    if result is None:
        return 0
    return result[0]


def _increment_callable_count(
        *,
        callable_: Union[Callable, str],
        module_name: str,
        class_: Optional[Type] = None) -> None:
    """
    Increment a specified callable count number.

    Parameters
    ----------
    callable_ : Callable or str
        Target function or method or property or dunder method name.
    module_name : str
        Module name. This value will be set the `__name__` value.
    class_ : Type or None, optional
        Target class type. If the target callable_ variable is not
        a method, this argument will be ignored.
    """
    from apysc._expression import expression_data_util
    callable_count: int = _get_callable_count(
        callable_=callable_, module_name=module_name, class_=class_)
    callable_count += 1
    path_name: str = _get_callable_path_name(
        callable_=callable_, module_name=module_name, class_=class_)
    table_name: str = \
        expression_data_util.TableName.DEBUG_MODE_CALLABLE_COUNT.value
    query: str = (
        f'DELETE FROM {table_name} '
        f"WHERE name = '{path_name}';"
    )
    expression_data_util.exec_query(sql=query, commit=False)
    query = (
        f'INSERT INTO {table_name}(name, count) '
        f"VALUES ('{path_name}', {callable_count});"
    )
    expression_data_util.exec_query(sql=query)


class DebugInfo:
    """
    Save a debug information (append callable interface name
    comment and arguments information) to the JavaScript
    expression file. This class is used at the `with` statement.
    """

    _callable: Union[Callable, str]
    _locals: Dict[str, Any]
    _module_name: str
    _class: Optional[Type]
    _DIVIDER: str = '/' * 70
    _callable_count: int
    _indent: Indent

    def __init__(
            self, callable_: Union[Callable, str],
            locals_: Dict[str, Any],
            module_name: str,
            *,
            class_: Optional[Type] = None) -> None:
        """
        Save a debug information (append callable interface name
        comment and arguments information) to the JavaScript
        expression file. This class is used at the `with` statement.

        Notes
        -----
        If the debug mode setting is not enabled, saving will
        be skipped.

        Parameters
        ----------
        callable_ : Callable or str
            Target function or method or property or dunder method name.
        locals_ : dict
            Local variables. This value will be set by the `locals()`
            function.
        module_name : str
            Module name. This value will be set the `__name__` value.
        class_ : Type or None, optional
            Target class type. If the target callable_ variable is not
            a method, this argument will be ignored.
        """
        self._callable = callable_
        self._locals = locals_
        self._module_name = module_name
        self._class = class_
        _increment_callable_count(
            callable_=callable_, module_name=module_name, class_=class_)
        self._callable_count = _get_callable_count(
            callable_=callable_, module_name=module_name, class_=class_)
        self._indent = Indent()

    def _get_class_info(self) -> str:
        """
        Get a class information string.

        Returns
        -------
        class_info : str
            Target class information string.
        """
        if self._class is None:
            class_info: str = ''
        else:
            class_info = f'\n// class: {self._class.__name__}'
        return class_info

    def __enter__(self) -> None:
        """
        The method will be called at the start of the with block.
        """
        import apysc as ap
        from apysc._type.variable_name_interface import VariableNameInterface
        if not ap.is_debug_mode():
            return
        class_info: str = self._get_class_info()
        if not self._locals:
            arguments_info: str = ''
        else:
            arguments_info = '\n// arguments and variables:'
            for argument_name, argument in self._locals.items():
                if inspect.ismodule(argument):
                    continue
                if argument_name.startswith('__'):
                    continue
                if isinstance(argument, str):
                    argument = repr(argument)[1:-1]
                    argument = f"'{argument}'"
                if isinstance(argument, VariableNameInterface):
                    argument = f'{argument}({argument.variable_name})'
                arguments_info += f'\n//    {argument_name} = {argument}'
        callable_str: str = _get_callable_str(callable_=self._callable)
        expression: str = (
            f'{self._DIVIDER}'
            f'\n// [{callable_str} {self._callable_count}] '
            'started.'
            f'\n// module name: {self._module_name}'
            f'{class_info}'
            f'{arguments_info}'
        )
        ap.append_js_expression(expression=expression)
        self._indent.__enter__()

    def __exit__(self, *args: Any) -> None:
        """
        The method will be called at the end of the with block.

        Parameters
        ----------
        *args : list
            Positional arguments.
        """
        import apysc as ap
        if not ap.is_debug_mode():
            return
        class_info: str = self._get_class_info()
        callable_str: str = _get_callable_str(callable_=self._callable)
        expression: str = (
            f'// [{callable_str} {self._callable_count}] ended.'
            f'\n// module name: {self._module_name}'
            f'{class_info}'
            f'\n{self._DIVIDER}'
        )
        self._indent.__exit__()
        ap.append_js_expression(expression=expression)



"""Path data class implementation for the svg's `close path` (Z).
"""

from apysc._geom.path_data_base import PathDataBase
from apysc._type.string import String


class PathClose(PathDataBase):
    """
    Path data class for the svg's `close path` (Z).
    """

    def __init__(self) -> None:
        """
        Path data class for the svg's `close path` (Z).
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='__init__', locals_=locals(),
                module_name=__name__, class_=PathClose):
            from apysc._geom.path_label import PathLabel
            super(PathClose, self).__init__(
                path_label=PathLabel.CLOSE,
                relative=False)

    def _get_svg_str(self) -> str:
        """
        Get a path's SVG string created with the current setting.

        Returns
        -------
        svg_str : str
            A path's SVG string created with the current setting.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self._get_svg_str, locals_=locals(),
                module_name=__name__, class_=PathClose):
            from apysc._type import value_util
            svg_char: String = self._get_svg_char()
            svg_str: str = value_util.get_value_str_for_expression(
                value=svg_char)
            return svg_str



"""Class implementation for mouse wheel event.
"""

from apysc._event.event import Event
from apysc._type.int import Int
from apysc._type.variable_name_interface import VariableNameInterface


class WheelEvent(Event):
    """
    Mouse wheel event class.
    """

    def __init__(self, this: VariableNameInterface) -> None:
        """
        Mouse wheel event class.

        Parameters
        ----------
        this : VariableNameInterface
            Instance will be binded this event.

        Notes
        -----
        Not supported each SVG elements' mouse wheel event currently, only
        supported document (overall screen) mouse wheel.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='__init__', locals_=locals(),
                module_name=__name__, class_=WheelEvent):
            from apysc._expression import var_names
            super(WheelEvent, self).__init__(
                this=this,
                type_name=var_names.WHEEL_EVENT)

    @property
    def delta_x(self) -> Int:
        """
        Horizontal mouse wheel value.

        Returns
        -------
        delta_x : Int
            Delta x value.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='delta_x', locals_=locals(),
                module_name=__name__, class_=WheelEvent):
            delta_x: ap.Int = ap.Int(0)
            self._append_delta_x_getter_expression(delta_x=delta_x)
            return delta_x

    def _append_delta_x_getter_expression(self, *, delta_x: Int) -> None:
        """
        Append delta_x getter property's expression.

        Parameters
        ----------
        delta_x : Int
            Target delta x value.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self._append_delta_x_getter_expression,
                locals_=locals(),
                module_name=__name__, class_=WheelEvent):
            expression: str = (
                f'{delta_x.variable_name} = '
                f'{self.variable_name}.deltaX;'
            )
            ap.append_js_expression(expression=expression)

    @property
    def delta_y(self) -> Int:
        """
        Vertical mouse wheel value.

        Returns
        -------
        delta_y : Int
            Delta y value.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='delta_y', locals_=locals(),
                module_name=__name__, class_=WheelEvent):
            delta_y: ap.Int = ap.Int(0)
            self._append_delta_y_getter_expression(delta_y=delta_y)
            return delta_y

    def _append_delta_y_getter_expression(self, *, delta_y: Int) -> None:
        """
        Append delta_y getter property's expression.

        Parameters
        ----------
        delta_y : Int
            Target delta y value.
        """
        import apysc as ap
        expression: str = (
            f'{delta_y.variable_name} = '
            f'{self.variable_name}.deltaY;'
        )
        ap.append_js_expression(expression=expression)



"""Path's validation implementations.
"""

from typing import List

from apysc._geom.path_bezier_2d import PathBezier2D
from apysc._geom.path_bezier_2d_continual import PathBezier2DContinual
from apysc._geom.path_bezier_3d import PathBezier3D
from apysc._geom.path_bezier_3d_continual import PathBezier3DContinual
from apysc._geom.path_data_base import PathDataBase


def validate_path_data_list(
        path_data_list: List[PathDataBase]) -> None:
    """
    Validate a specified path data list.

    Parameters
    ----------
    path_data_list : list of PathDataBase
        Target path data settings, such as the ap.PathData.MoveTo.

    Raises
    ------
    ValueError
        - If a specified path data list is empty.
        - If a preceding data of `PathBezier2DContinual` instance is
            not a `PathBezier2D` or `PathBezier2DContinual` one.
        - If a preceding data of `PathBezier3DContinual` instance is
            not a `PathBezier3D` or `PathBezier3DContinual` one.
    """
    if not path_data_list:
        raise ValueError(
            '`path_data_list` argument can not be empty.')
    _validate_bezier_2d_continual_pre_data(path_data_list=path_data_list)
    _validate_bezier_3d_continual_pre_data(path_data_list=path_data_list)


def _validate_bezier_3d_continual_pre_data(
        path_data_list: List[PathDataBase]) -> None:
    """
    Validate a preceding data of `PathBezier3DContinual` instance
    in list is a `PathBezier3D` or `PathBezier3DContinual`.

    Parameters
    ----------
    path_data_list : list of PathDataBase
        Target path data settings, such as the ap.PathData.MoveTo.

    Raises
    ------
    ValueError
        If a preceding data of `PathBezier3DContinual` instance is
        not a `PathBezier3D` or `PathBezier3DContinual` one.
    """
    for i, path_data in enumerate(path_data_list):
        if i == 0 and isinstance(path_data, PathBezier3DContinual):
            raise ValueError(
                '`PathBezier3DContinual` instance can not be set '
                'at the first position of `path_data_list`.')
        if i == 0:
            continue
        if not isinstance(path_data, PathBezier3DContinual):
            continue
        pre_data: PathDataBase = path_data_list[i - 1]
        if (not isinstance(pre_data, PathBezier3D)
                and not isinstance(pre_data, PathBezier3DContinual)):
            raise ValueError(
                '`PathBezier3DContinual` instance in a '
                '`path_data_list` argument can set only after a '
                '`PathBezier3D` or `PathBezier3DContinual` one.'
                f'\nActual: {type(pre_data)}'
                f'\nIndex: {i}')


def _validate_bezier_2d_continual_pre_data(
        path_data_list: List[PathDataBase]) -> None:
    """
    Validate a preceding data of `PathBezier2DContinual` instance
    in list is a `PathBezier2D` or `PathBezier2DContinual`.

    Parameters
    ----------
    path_data_list : list of PathDataBase
        Target path data settings, such as the ap.PathData.MoveTo.

    Raises
    ------
    ValueError
        If a preceding data of `PathBezier2DContinual` instance is
        not a `PathBezier2D` or `PathBezier2DContinual` one.
    """
    for i, path_data in enumerate(path_data_list):
        if i == 0 and isinstance(path_data, PathBezier2DContinual):
            raise ValueError(
                '`PathBezier2DContinual` instance can not be set '
                'at the first position of `path_data_list`.')
        if i == 0:
            continue
        if not isinstance(path_data, PathBezier2DContinual):
            continue
        pre_data: PathDataBase = path_data_list[i - 1]
        if (not isinstance(pre_data, PathBezier2D)
                and not isinstance(pre_data, PathBezier2DContinual)):
            raise ValueError(
                '`PathBezier2DContinual` instance in a '
                '`path_data_list` argument can set only after a '
                '`PathBezier2D` or `PathBezier2DContinual` one.'
                f'\nActual: {type(pre_data)}'
                f'\nIndex: {i}')



"""For loop class implementation.
"""

from typing import Any
from typing import Dict
from typing import Generic
from typing import Optional
from typing import TypeVar
from typing import Union

from apysc._expression.indent_num import Indent
from apysc._type.array import Array
from apysc._type.dictionary import Dictionary
from apysc._type.int import Int
from apysc._type.number import Number
from apysc._type.string import String

T = TypeVar('T', Int, String, Number)


class For(Generic[T]):
    """
    A class to append for the (loop) expression.

    References
    ----------
    - For document
        - https://simon-ritchie.github.io/apysc/for.html
    """

    _arr_or_dict: Union[Array, Dictionary]
    _locals: Dict[str, Any]
    _globals: Dict[str, Any]
    _snapshot_name: str
    _indent: Indent

    def __init__(
            self, arr_or_dict: Union[Array, Dictionary],
            *,
            locals_: Optional[Dict[str, Any]] = None,
            globals_: Optional[Dict[str, Any]] = None) -> None:
        """
        A class to append for the (loop) expression.

        Parameters
        ----------
        arr_or_dict : Array or Dictionary
            Array or Dictionary instance to iterate.
        locals_ : dict or None, default None
            Current scope's local variables. Set locals() value to
            this argument. If specified, all local scope
            VariableNameInterface variables (like Int, Sprite) will be
            reverted ad the end of For scope. This setting is useful
            when you don't want to update each variable by the
            implementation of the For scope.
        globals_ : dict or None, default None
            Current scope's golobal variables. Set golobals() value
            to this argument. If specified, all local scope
            VariableNameInterface variables (like Int, Sprite) will be
            reverted ad the end of For scope. This setting is useful
            when you don't want to update each variable by the
            implementation of the For scope.

        References
        ----------
        - For document
            - https://simon-ritchie.github.io/apysc/for.html

        Examples
        --------
        >>> import apysc as ap
        >>> arr: ap.Array = ap.Array(range(3))
        >>> with ap.For(arr) as i:
        ...     ap.trace('Loop index is:', i)
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='__init__', locals_=locals(),
                module_name=__name__, class_=For):
            self._validate_arr_or_dict_val_type(arr_or_dict=arr_or_dict)
            if locals_ is None:
                locals_ = {}
            if globals_ is None:
                globals_ = {}
            self._arr_or_dict = arr_or_dict
            self._locals = locals_
            self._globals = globals_
            self._indent = Indent()

    def _validate_arr_or_dict_val_type(
            self, *, arr_or_dict: Union[Array, Dictionary]) -> None:
        """
        Validate loop value type is Array of Dictionary.

        Parameters
        ----------
        arr_or_dict : Array or Dictionary
            Value to be checked.

        Raises
        ------
        TypeError
            If value type is neither Array nor Dictionary.
        """
        if isinstance(arr_or_dict, (Array, Dictionary)):
            return
        raise TypeError(
            'Specified value type is neither Array nor Dictionary: '
            f'{type(arr_or_dict)}')

    def __enter__(self) -> T:
        """
        Method to be called when begining of with statement.

        Returns
        -------
        i_or_key : Int or String
            Loop index or dictionary key.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='__enter__', locals_=locals(),
                module_name=__name__, class_=For):
            from apysc._loop import loop_count
            from apysc._type import revert_interface
            loop_count.increment_current_loop_count()
            self._snapshot_name = \
                revert_interface.make_snapshots_of_each_scope_vars(
                    locals_=self._locals, globals_=self._globals)
            i_or_key: Union[ap.Int, ap.String]
            if isinstance(self._arr_or_dict, ap.Array):
                i_or_key = ap.Int(0)
                self._append_arr_enter_expression(i=i_or_key)
            else:
                if self._arr_or_dict._value:
                    key: str = str(list(self._arr_or_dict._value.keys())[0])
                else:
                    key = ''
                i_or_key = ap.String(key)
                self._append_dict_enter_expression(key=i_or_key)
            self._indent.__enter__()
            return i_or_key  # type: ignore

    def __exit__(self, *args: Any) -> None:
        """
        Method to be called when end of with statement.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='__exit__', locals_=locals(),
                module_name=__name__, class_=For):
            from apysc._expression import last_scope
            from apysc._expression.last_scope import LastScope
            from apysc._loop import loop_count
            from apysc._type import revert_interface
            loop_count.decrement_current_loop_count()
            revert_interface.revert_each_scope_vars(
                snapshot_name=self._snapshot_name,
                locals_=self._locals, globals_=self._globals)
            self._indent.__exit__()
            ap.append_js_expression(expression='}')
            last_scope.set_last_scope(value=LastScope.FOR)

    def _append_arr_enter_expression(self, *, i: Int) -> None:
        """
        Append for loop start expression (for Array value).

        Parameters
        ----------
        i : Int
            Loop index value.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self._append_arr_enter_expression, locals_=locals(),
                module_name=__name__, class_=For):
            i_name: str = i.variable_name
            expression: str = (
                f'var length = {self._arr_or_dict.variable_name}.length;\n'
                f'for ({i_name} = 0; {i_name} < length; {i_name}++) {{'
            )
            ap.append_js_expression(expression=expression)

    def _append_dict_enter_expression(self, *, key: String) -> None:
        """
        Append for loop start expression (for Dictionary value).

        Parameters
        ----------
        key : String
            Loop (dictionary) key value.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self._append_dict_enter_expression,
                locals_=locals(),
                module_name=__name__, class_=For):
            key_name: str = key.variable_name
            expression: str = (
                f'for (var {key_name} in '
                f'{self._arr_or_dict.variable_name}) {{'
            )
            ap.append_js_expression(expression=expression)



"""Class implementation for mouse up interface.
"""

from typing import Callable
from typing import Dict
from typing import Optional
from typing import TypeVar

from apysc._event.handler import HandlerData
from apysc._event.mouse_event import MouseEvent
from apysc._event.mouse_event_interface_base import MouseEventInterfaceBase

_O = TypeVar('_O')
_Handler = Callable[[MouseEvent, _O], None]


class MouseUpInterface(MouseEventInterfaceBase):

    _mouse_up_handlers: Dict[str, HandlerData]

    def mouseup(
            self, handler: _Handler[_O], *,
            options: Optional[_O] = None) -> str:
        """
        Add mouse up event listener setting.

        Parameters
        ----------
        handler : _Handler
            Callable that called when mouse is upped on this instance.
        options : dict or None, default None
            Optional arguments dictionary to be passed to handler.

        Returns
        -------
        name : str
            Handler's name.

        References
        ----------
        - Mousedown and mouseup interfaces document
            - https://bit.ly/3zgjOnF
        - About the handler options’ type document
            - https://bit.ly/39tnYxC
        """
        import apysc as ap
        from apysc._validation.variable_name_validation import \
            validate_variable_name_interface_type
        with ap.DebugInfo(
                callable_=self.mouseup, locals_=locals(),
                module_name=__name__, class_=MouseUpInterface):
            from apysc._event.handler import append_handler_expression
            from apysc._event.handler import get_handler_name
            from apysc._type.variable_name_interface import \
                VariableNameInterface
            from apysc._validation.handler_options_validation import \
                validate_options_type
            self_instance: VariableNameInterface = \
                validate_variable_name_interface_type(instance=self)
            validate_options_type(options=options)
            self._initialize_mouse_up_handlers_if_not_initialized()
            name: str = get_handler_name(handler=handler, instance=self)
            self._set_mouse_event_handler_data(
                handler=handler, handlers_dict=self._mouse_up_handlers,
                options=options)
            self._append_mouse_event_binding_expression(
                name=name, mouse_event_type=ap.MouseEventType.MOUSEUP)
            e: ap.MouseEvent = ap.MouseEvent(this=self_instance)
            append_handler_expression(
                handler_data=self._mouse_up_handlers[name],
                handler_name=name, e=e)
            return name

    def _initialize_mouse_up_handlers_if_not_initialized(self) -> None:
        """
        Initialize _mouse_up_handlers attribute if it is not
        initialized yet.
        """
        if hasattr(self, '_mouse_up_handlers'):
            return
        self._mouse_up_handlers = {}

    def unbind_mouseup(self, handler: _Handler[_O]) -> None:
        """
        Unbind specified handler's mouse up event.

        Parameters
        ----------
        handler : _Handler
            Callable to be unbinded.

        References
        ----------
        - Mousedown and mouseup interfaces document
            - https://bit.ly/3zgjOnF
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self.unbind_mouseup, locals_=locals(),
                module_name=__name__, class_=MouseUpInterface):
            self._initialize_mouse_up_handlers_if_not_initialized()
            self._unbind_mouse_event(
                handler=handler, mouse_event_type=ap.MouseEventType.MOUSEUP,
                handlers_dict=self._mouse_up_handlers)

    def unbind_mouseup_all(self) -> None:
        """
        Unbind all mouse up events.

        References
        ----------
        - Mousedown and mouseup interfaces document
            - https://bit.ly/3zgjOnF
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self.unbind_mouseup_all, locals_=locals(),
                module_name=__name__, class_=MouseUpInterface):
            self._initialize_mouse_up_handlers_if_not_initialized()
            self._unbind_all_mouse_events(
                mouse_event_type=ap.MouseEventType.MOUSEUP,
                handlers_dict=self._mouse_up_handlers)



"""Document (overall window) class implementation.
"""

from apysc._type.variable_name_interface import VariableNameInterface


class Document(VariableNameInterface):
    """
    Document (overall HTML document) class.
    """

    def __init__(self) -> None:
        """
        Document (overall HTML document) class.
        """
        from apysc._expression import var_names
        self.variable_name = var_names.DOCUMENT


document: Document = Document()



"""Implementations of expression's indent number related interfaces.

Mainly following interfaces are defined:

- get_current_indent_num: Get a current indent number.
- Indent: Class implementation for increment and decrement
    indentation number. Basically use this class at with statement.
- reset: Reset current indent number.
"""

from typing import Any
from typing import List
from typing import Optional
from typing import Tuple


def get_current_indent_num() -> int:
    """
    Get a current indent number.

    Returns
    -------
    current_indent_num : int
        Current indent number.
    """
    from apysc._expression import expression_data_util
    table_name: str = _get_indent_num_table_name()
    query: str = (
        f'SELECT num FROM {table_name} LIMIT 1;'
    )
    expression_data_util.exec_query(sql=query)
    result: Optional[Tuple[int]] = expression_data_util.cursor.fetchone()
    if result is None:
        return 0
    current_indent_num: int = result[0]
    return current_indent_num


def _save_current_indent_num(*, indent_num: int) -> None:
    """
    Save the current indentation number.

    Parameters
    ----------
    indent_num : int
        Current indentation number.
    """
    from apysc._expression import expression_data_util
    table_name: str = _get_indent_num_table_name()
    query: str = f'DELETE FROM {table_name};'
    expression_data_util.exec_query(sql=query, commit=False)
    query = (
        f'INSERT INTO {table_name}(num) VALUES ({indent_num});'
    )
    expression_data_util.exec_query(sql=query)


def _get_indent_num_table_name() -> str:
    """
    Get a indentation number table name. This value will switch
    by scope condition (e.g., event handler's scope or not).

    Returns
    -------
    table_name : str
        Target table name.
    """
    from apysc._expression import event_handler_scope
    from apysc._expression import expression_data_util
    event_handler_scope_count: int = \
        event_handler_scope.get_current_event_handler_scope_count()
    if event_handler_scope_count == 0:
        return expression_data_util.TableName.INDENT_NUM_NORMAL.value
    return expression_data_util.TableName.INDENT_NUM_HANDLER.value


class Indent:
    """Class implementation for increment and decrement
    indentation number. Basically use this class at with statement.
    """

    def __enter__(self) -> None:
        """
        Method to be used by with statement.
        This method will increment indentation number.
        """
        current_indent_num: int = get_current_indent_num()
        current_indent_num += 1
        _save_current_indent_num(indent_num=current_indent_num)

    def __exit__(self, *args: Any) -> None:
        """
        Method to be used by with statement.
        This method will decrement indentation number.

        Parameters
        ----------
        *args : list
            Any positional arguments.
        """
        current_indent_num: int = get_current_indent_num()
        current_indent_num -= 1
        if current_indent_num < 0:
            current_indent_num = 0
        _save_current_indent_num(indent_num=current_indent_num)


def reset() -> None:
    """
    Reset current indent number.
    """
    from apysc._expression import expression_data_util
    table_names: List[str] = [
        expression_data_util.TableName.INDENT_NUM_NORMAL.value,
        expression_data_util.TableName.INDENT_NUM_HANDLER.value,
    ]
    for table_name in table_names:
        query: str = f'DELETE FROM {table_name};'
        expression_data_util.exec_query(sql=query, commit=False)
    expression_data_util.connection.commit()



"""Dash setting class implementation for line.
"""

from typing import Union

from apysc._type.dictionary import Dictionary
from apysc._type.int import Int


class LineDashSetting(Dictionary[str, Int]):
    """
    Dash setting class for a line.

    References
    ----------
    - Graphics line_style interface document
        - https://bit.ly/3zauILT
    """

    def __init__(
            self, dash_size: Union[int, Int],
            space_size: Union[int, Int]) -> None:
        """
        Dash setting class for a line.

        Parameters
        ----------
        dash_size : int or Int
            Dash size.
        space_size : int or Int
            Blank space size between dashes.

        References
        ----------
        - Graphics line_style interface document
            - https://bit.ly/3zauILT
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='__init__', locals_=locals(),
                module_name=__name__, class_=LineDashSetting):
            from apysc._converter.to_apysc_val_from_builtin import \
                get_copied_int_from_builtin_val
            from apysc._validation import number_validation
            number_validation.validate_nums_are_int_and_gt_zero(
                nums=[dash_size, space_size])
            dash_size_: ap.Int = get_copied_int_from_builtin_val(
                integer=dash_size)
            space_size_: ap.Int = get_copied_int_from_builtin_val(
                integer=space_size)
            super(LineDashSetting, self).__init__({
                'dash_size': dash_size_,
                'space_size': space_size_,
            })

    @property
    def dash_size(self) -> Int:
        """
        Get a dash size setting.

        Returns
        -------
        dash_size : Int
            Dash size setting.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='dash_size', locals_=locals(),
                module_name=__name__, class_=LineDashSetting):
            return self['dash_size']

    @property
    def space_size(self) -> Int:
        """
        Get a space size setting.

        Returns
        -------
        space_size : Int
            Space size setting.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='space_size', locals_=locals(),
                module_name=__name__, class_=LineDashSetting):
            return self['space_size']



"""Class implementation for the prevent_default interface.
"""

from apysc._type.variable_name_interface import VariableNameInterface


class PreventDefaultInterface(VariableNameInterface):

    def prevent_default(self) -> None:
        """
        Prevent event's default behavior.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self.prevent_default, locals_=locals(),
                module_name=__name__, class_=PreventDefaultInterface):
            expression: str = (
                f'{self.variable_name}.preventDefault();'
            )
            ap.append_js_expression(expression=expression)



"""Class implementation for line dot setting interface.
"""

from typing import Dict
from typing import Optional

from apysc._display.line_dot_setting import LineDotSetting
from apysc._type.revert_interface import RevertInterface
from apysc._type.variable_name_interface import VariableNameInterface


class LineDotSettingInterface(VariableNameInterface, RevertInterface):

    _line_dot_setting: Optional[LineDotSetting]

    def _initialize_line_dot_setting_if_not_initialized(self) -> None:
        """
        Initialize _line_dot_setting attribute if it is not
        initialized yet.
        """
        if hasattr(self, '_line_dot_setting'):
            return
        self._line_dot_setting = None

    @property
    def line_dot_setting(self) -> Optional[LineDotSetting]:
        """
        Get this instance's line dot setting.

        Returns
        -------
        line_dot_setting : LineDotSetting or None
            Lien dot setting.

        References
        ----------
        - Graphics line_dot_setting interface document
            - https://bit.ly/3ni8rIS
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='line_dot_setting', locals_=locals(),
                module_name=__name__, class_=LineDotSettingInterface):
            self._initialize_line_dot_setting_if_not_initialized()
            return self._line_dot_setting

    @line_dot_setting.setter
    def line_dot_setting(self, value: Optional[LineDotSetting]) -> None:
        """
        Set line dot setting.

        Parameters
        ----------
        value : LineDotSetting or None
            Line dot setting to set.

        References
        ----------
        - Graphics line_dot_setting interface document
            - https://bit.ly/3ni8rIS
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=LineDotSettingInterface, locals_=locals(),
                module_name=__name__, class_=LineDotSettingInterface):
            from apysc._validation import display_validation
            self._update_line_dot_setting_and_skip_appending_exp(value=value)
            self._append_line_dot_setting_update_expression()
            display_validation.validate_multiple_line_settings_isnt_set(
                any_instance=self)

    def _update_line_dot_setting_and_skip_appending_exp(
            self, *, value: Optional[LineDotSetting]) -> None:
        """
        Update line dot setting and skip appending expression.

        Parameters
        ----------
        value : LineDotSetting or None
            Line dot setting to set.
        """
        if value is not None and not isinstance(value, LineDotSetting):
            raise TypeError(
                'Not supported line_dot_setting type specified: '
                f'{type(value)}'
                '\nAcceptable ones are: LineDotSetting or None.')
        self._line_dot_setting = value

    def _append_line_dot_setting_update_expression(self) -> None:
        """
        Append line dot setting updating expression.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self._append_line_dot_setting_update_expression,
                locals_=locals(),
                module_name=__name__, class_=LineDotSettingInterface):
            if self._line_dot_setting is None:
                setting_str: str = '""'
            else:
                setting_str = self._line_dot_setting.dot_size.variable_name
            expression: str = (
                f'{self.variable_name}.css("stroke-dasharray", {setting_str});'
            )
            ap.append_js_expression(expression=expression)

    _line_dot_setting_snapshots: Dict[str, Optional[LineDotSetting]]

    def _make_snapshot(self, *, snapshot_name: str) -> None:
        """
        Make value's snapshot.

        Parameters
        ----------
        snapshot_name : str
            Target snapshot name.
        """
        self._initialize_line_dot_setting_if_not_initialized()
        self._set_single_snapshot_val_to_dict(
            dict_name='_line_dot_setting_snapshots',
            value=self._line_dot_setting, snapshot_name=snapshot_name)

    def _revert(self, *, snapshot_name: str) -> None:
        """
        Revert value if snapshot exists.

        Parameters
        ----------
        snapshot_name : str
            Target snapshot name.
        """
        if not self._snapshot_exists(snapshot_name=snapshot_name):
            return
        self._line_dot_setting = self._line_dot_setting_snapshots[
            snapshot_name]



"""Implementation of mouse wheel event interfaces.

Notes
-----
Not supported each SVG elements' mouse wheel event currently, only
supported document (overall screen) mouse wheel.
"""

from typing import Callable
from typing import Optional
from typing import TypeVar

from apysc._event.wheel_event import WheelEvent

_O = TypeVar('_O')
_Handler = Callable[[WheelEvent, _O], None]


def bind_wheel_event_to_document(
        handler: _Handler[_O],
        *,
        options: Optional[_O] = None) -> str:
    """
    Bind wheel event to document (overall window).

    Parameters
    ----------
    handler : _Handler
        Callable that handle wheel event.
    options : dict or None, default None
        Optional arguments dictionary to pass to.

    Returns
    -------
    name : str
        Handler's name.

    References
    ----------
    - About the handler options’ type document
        - https://bit.ly/39tnYxC
    """
    import apysc as ap
    with ap.DebugInfo(
            callable_=bind_wheel_event_to_document, locals_=locals(),
            module_name=__name__):
        from apysc._event.handler import HandlerData
        from apysc._event.handler import append_handler_expression
        from apysc._event.handler import get_handler_name
        from apysc._validation.handler_options_validation import \
            validate_options_type
        validate_options_type(options=options)
        name: str = get_handler_name(handler=handler, instance=ap.document)
        expression: str = (
            f'$({ap.document.variable_name}).on("mousewheel", {name});'
        )
        ap.append_js_expression(expression=expression)

        if options is None:
            options = {}  # type: ignore
        handler_data: HandlerData = {  # type: ignore
            'handler': handler,  # type: ignore
            'options': options,
        }
        e: ap.WheelEvent = ap.WheelEvent(this=ap.document)
        append_handler_expression(
            handler_data=handler_data, handler_name=name, e=e)
        return name


def unbind_wheel_event_from_document(handler: _Handler[_O]) -> None:
    """
    Unbind specified handler's wheel event from document (overall window).

    Parameters
    ----------
    handler : _Handler
        Callable to unbind.
    """
    import apysc as ap
    with ap.DebugInfo(
            callable_=unbind_wheel_event_from_document, locals_=locals(),
            module_name=__name__):
        from apysc._event.handler import get_handler_name
        name: str = get_handler_name(handler=handler, instance=ap.document)
        expression: str = (
            f'$({ap.document.variable_name}).off("mousewheel", {name});'
        )
        ap.append_js_expression(expression=expression)


def unbind_wheel_event_all_from_document() -> None:
    """
    Unbind all wheel event from document (overall window).
    """
    import apysc as ap
    with ap.DebugInfo(
            callable_=unbind_wheel_event_all_from_document, locals_=locals(),
            module_name=__name__):
        expression: str = (
            f'$({ap.document.variable_name}).off("mousewheel");'
        )
        ap.append_js_expression(expression=expression)



"""Base class implementation for each lines.
"""

from abc import ABC
from abc import abstractmethod

from apysc._display import graphics
from apysc._display.fill_alpha_interface import FillAlphaInterface
from apysc._display.fill_color_interface import FillColorInterface
from apysc._display.graphics_base import GraphicsBase
from apysc._display.line_alpha_interface import LineAlphaInterface
from apysc._display.line_color_interface import LineColorInterface
from apysc._display.line_dash_dot_setting_interface import \
    LineDashDotSettingInterface
from apysc._display.line_dash_setting_interface import LineDashSettingInterface
from apysc._display.line_dot_setting_interface import LineDotSettingInterface
from apysc._display.line_joints_interface import LineJointsInterface
from apysc._display.line_round_dot_setting_interface import \
    LineRoundDotSettingInterface


class LineBase(
        GraphicsBase, FillColorInterface, FillAlphaInterface,
        LineColorInterface, LineAlphaInterface,
        LineJointsInterface, LineDotSettingInterface,
        LineDashSettingInterface, LineRoundDotSettingInterface,
        LineDashDotSettingInterface, ABC):

    @abstractmethod
    def __repr__(self) -> str:
        """
        Get a string representation of this instance (for the sake of
        debugging).
        """

    def _set_line_setting_if_not_none_value_exists(
            self, *, parent_graphics: 'graphics.Graphics') -> None:
        """
        If a line setting (dot, dash, or something else) with a value
        other than None exists, set that value to the attribute.

        Parameters
        ----------
        parent_graphics : Graphics
            Parent Graphics instance.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self._set_line_setting_if_not_none_value_exists,
                locals_=locals(),
                module_name=__name__, class_=LineBase):
            from apysc._display.graphics import Graphics
            parent_graphics_: Graphics = parent_graphics
            if parent_graphics_.line_dot_setting is not None:
                self.line_dot_setting = parent_graphics_.line_dot_setting
                return
            if parent_graphics_.line_dash_setting is not None:
                self.line_dash_setting = parent_graphics_.line_dash_setting
                return
            if parent_graphics_.line_round_dot_setting is not None:
                self.line_round_dot_setting = \
                    parent_graphics_.line_round_dot_setting
                return
            if parent_graphics_.line_dash_dot_setting is not None:
                self.line_dash_dot_setting = \
                    parent_graphics_.line_dash_dot_setting
                return

    def _set_initial_basic_values(
            self, *, parent: 'graphics.Graphics') -> None:
        """
        Set initial basic values (fill color, line thickness, and so on).

        Parameters
        ----------
        parent : Graphics
            Graphics instance to link this graphic.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self._set_initial_basic_values, locals_=locals(),
                module_name=__name__, class_=LineBase):
            from apysc._display.graphics import Graphics
            parent_graphics: Graphics = parent
            self._set_initial_fill_color_if_not_blank(
                fill_color=parent_graphics.fill_color)
            self._update_fill_alpha_and_skip_appending_exp(
                value=parent_graphics.fill_alpha)
            self._set_initial_line_color_if_not_blank(
                line_color=parent_graphics.line_color)
            self._update_line_thickness_and_skip_appending_exp(
                value=parent_graphics.line_thickness)
            self._update_line_alpha_and_skip_appending_exp(
                value=parent_graphics.line_alpha)
            self._initialize_x_if_not_initialized()
            self._initialize_y_if_not_initialized()
            self._update_line_cap_and_skip_appending_exp(
                value=parent_graphics.line_cap)
            self._update_line_joints_and_skip_appending_exp(
                value=parent_graphics.line_joints)

            self._append_applying_new_attr_val_exp(
                new_attr=self._fill_alpha, attr_name='fill_alpha')
            self._append_attr_to_linking_stack(
                attr=self._fill_alpha, attr_name='fill_alpha')

            self._append_applying_new_attr_val_exp(
                new_attr=self._fill_color, attr_name='fill_color')
            self._append_attr_to_linking_stack(
                attr=self._fill_color, attr_name='fill_color')

            self._append_applying_new_attr_val_exp(
                new_attr=self._line_color, attr_name='line_color')
            self._append_attr_to_linking_stack(
                attr=self._line_color, attr_name='line_color')

            self._append_applying_new_attr_val_exp(
                new_attr=self._line_alpha, attr_name='line_alpha')
            self._append_attr_to_linking_stack(
                attr=self._line_alpha, attr_name='line_alpha')

            self._append_applying_new_attr_val_exp(
                new_attr=self._line_thickness, attr_name='line_thickness')
            self._append_attr_to_linking_stack(
                attr=self._line_thickness, attr_name='line_thickness')

    def _append_basic_vals_expression(
            self, *, expression: str, indent_num: int) -> str:
        """
        Append basic values expression to specified one.

        Parameters
        ----------
        expression : str
            Target expression.
        indent_num : int
            Indentation number.

        Returns
        -------
        expression : str
            After appending expression.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self._append_basic_vals_expression, locals_=locals(),
                module_name=__name__, class_=LineBase):
            from apysc._display import graphics_expression
            from apysc._display.graphics import Graphics
            graphics: Graphics = self.parent_graphics
            expression = graphics_expression.append_fill_expression(
                graphics=graphics, expression=expression,
                indent_num=indent_num)
            expression = graphics_expression.append_fill_opacity_expression(
                graphics=graphics, expression=expression,
                indent_num=indent_num)
            expression = graphics_expression.append_stroke_expression(
                graphics=graphics, expression=expression,
                indent_num=indent_num)
            expression = graphics_expression.append_stroke_width_expression(
                graphics=graphics, expression=expression,
                indent_num=indent_num)
            expression = graphics_expression.append_stroke_opacity_expression(
                graphics=graphics, expression=expression,
                indent_num=indent_num)
            expression = graphics_expression.append_stroke_linecap_expression(
                graphics=graphics, expression=expression,
                indent_num=indent_num)
            expression = graphics_expression.append_stroke_linejoin_expression(
                graphics=graphics, expression=expression,
                indent_num=indent_num)
            expression = graphics_expression.append_x_expression(
                graphic=self, expression=expression, indent_num=indent_num)
            expression = graphics_expression.append_y_expression(
                graphic=self, expression=expression, indent_num=indent_num)
            return expression



"""Variable name related Validation interfaces.
"""

from typing import Any

from apysc._type.variable_name_interface import VariableNameInterface


def validate_variable_name_interface_type(
        instance: Any) -> VariableNameInterface:
    """
    Validate specified instance type is VariableNameInterface.

    Parameters
    ----------
    instance : *
        Instance to be checked.

    Returns
    -------
    instance : VariableNameInterface
        Checked instance.

    Raises
    ------
    TypeError
        If specified instance type isn't VariableNameInterface.
    """
    if not isinstance(instance, VariableNameInterface):
        raise TypeError(
            'Specified instance type is not VariableNameInterface : '
            f'{type(instance)}')
    return instance



"""Else branch instruction implementation.
"""

from typing import Any
from typing import Dict
from typing import Optional

from apysc._branch.if_base import IfBase


class Else(IfBase):
    """
    A class to append else branch instruction expression.

    References
    ----------
    - Else document
        - https://simon-ritchie.github.io/apysc/else.html
    - Each branch instruction class scope variables reverting setting
        - https://bit.ly/3rkAuaT
    """

    def __init__(
            self,
            *,
            locals_: Optional[Dict[str, Any]] = None,
            globals_: Optional[Dict[str, Any]] = None) -> None:
        """
        A class to append else branch instruction expression.

        Notes
        -----
        - You can only use this class immediately after the
            `If` or `Elif` statement.

        Parameters
        ----------
        locals_ : dict or None, default None
            Current scope's local variables. Set locals() value to
            this argument.
        globals_ : dict or None, default None
            Current scope's golobal variables. Set golobals() value
            to this argument.

        References
        ----------
        - Else document
            - https://simon-ritchie.github.io/apysc/else.html
        - Each branch instruction class scope variables reverting setting
            - https://bit.ly/3rkAuaT

        Examples
        --------
        >>> import apysc as ap
        >>> int_val: ap.Int = ap.Int(10)
        >>> with ap.If(int_val >= 11):
        ...     ap.trace('Value is greater than equal 11.')
        >>> with ap.Else():
        ...     ap.trace('Value is less than 11.')
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='__init__', locals_=locals(),
                module_name=__name__, class_=Else):
            super().__init__(
                condition=None,
                locals_=locals_,
                globals_=globals_)

    def _append_enter_expression(self) -> None:
        """
        Append else branch instruction start expression.

        Raises
        ------
        ValueError
            If the last scope is not If or Elif.
        """
        import apysc as ap
        if not self._last_scope_is_if_or_elif():
            raise ValueError(
                'Else interface can only use right after If or Elif '
                'interfaces.')
        expression: str = (
            'else {'
        )
        ap.append_js_expression(expression=expression)

    def _set_last_scope(self) -> None:
        """
        Set expression last scope value.
        """
        from apysc._expression import last_scope
        from apysc._expression.last_scope import LastScope
        last_scope.set_last_scope(value=LastScope.ELSE)



"""Class implementation for line color interface.
"""

from typing import Dict
from typing import Union

from apysc._animation.animation_line_color_interface import \
    AnimationLineColorInterface
from apysc._type.attr_linking_interface import AttrLinkingInterface
from apysc._type.revert_interface import RevertInterface
from apysc._type.string import String


class LineColorInterface(
        AnimationLineColorInterface, RevertInterface, AttrLinkingInterface):

    _line_color: String

    @property
    def line_color(self) -> String:
        """
        Get this instance's line color.

        Returns
        -------
        line_color : String
            Current line color (hexadecimal string, e.g., '#00aaff').
            If not be set, blank string will be returned.

        References
        ----------
        - Graphics line_color interface document
            - https://simon-ritchie.github.io/apysc/graphics_line_color.html
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='line_color', locals_=locals(),
                module_name=__name__, class_=LineColorInterface):
            from apysc._type import value_util
            self._initialize_line_color_if_not_initialized()
            line_color: ap.String = value_util.get_copy(
                value=self._line_color)
            return line_color

    @line_color.setter
    def line_color(self, value: String) -> None:
        """
        Update this instance's line color.

        Parameters
        ----------
        value : str
            Line color to set.

        References
        ----------
        - Graphics line_color interface document
            - https://simon-ritchie.github.io/apysc/graphics_line_color.html
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='line_color', locals_=locals(),
                module_name=__name__, class_=LineColorInterface):
            self._initialize_line_color_if_not_initialized()
            self._update_line_color_and_skip_appending_exp(value=value)
            self._append_line_color_update_expression()

            self._append_applying_new_attr_val_exp(
                new_attr=self._line_color, attr_name='line_color')
            self._append_attr_to_linking_stack(
                attr=self._line_color, attr_name='line_color')

    def _append_line_color_update_expression(self) -> None:
        """
        Append line color updating expression.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self._append_line_color_update_expression,
                locals_=locals(),
                module_name=__name__, class_=LineColorInterface):
            from apysc._type import value_util
            line_color_str: str = value_util.get_value_str_for_expression(
                value=self._line_color)
            expression: str = (
                f'{self.variable_name}.stroke({line_color_str});'
            )
            ap.append_js_expression(expression=expression)

    def _set_initial_line_color_if_not_blank(
            self, *, line_color: Union[str, String]) -> None:
        """
        Set initial line color value if specified value is not
        blank string.

        Parameters
        ----------
        line_color : str or String
            Line color (hexadecimal string, e.g., '#00aaff').
        """
        self._initialize_line_color_if_not_initialized()
        if line_color == '':
            return
        if isinstance(line_color, str):
            line_color = String(line_color)
        self._update_line_color_and_skip_appending_exp(value=line_color)

    def _update_line_color_and_skip_appending_exp(
            self, *, value: String) -> None:
        """
        Update line color and skip appending expression.

        Parameters
        ----------
        value : String
            Line color to set.
        """
        from apysc._color import color_util
        self._initialize_line_color_if_not_initialized()
        value = color_util.complement_hex_color(
            hex_color_code=value)
        self._initialize_line_color_if_not_initialized()
        self._line_color = value

    def _initialize_line_color_if_not_initialized(self) -> None:
        """
        Initialize line_color attribute if that value is not
        initialized yet.
        """
        if hasattr(self, '_line_color'):
            return
        self._line_color = String('')

    _line_color_snapshots: Dict[str, str]

    def _make_snapshot(self, *, snapshot_name: str) -> None:
        """
        Make value's snapshot.

        Parameters
        ----------
        snapshot_name : str
            Target snapshot name.
        """
        self._initialize_line_color_if_not_initialized()
        self._set_single_snapshot_val_to_dict(
            dict_name='_line_color_snapshots',
            value=self._line_color._value, snapshot_name=snapshot_name)

    def _revert(self, *, snapshot_name: str) -> None:
        """
        Revert value if snapshot exists.

        Parameters
        ----------
        snapshot_name : str
            Target snapshot name.
        """
        if not self._snapshot_exists(snapshot_name=snapshot_name):
            return
        self._line_color._value = self._line_color_snapshots[snapshot_name]



"""Implementations of the Line class.
"""


from apysc._display import graphics
from apysc._display.line_base import LineBase
from apysc._geom import point2d


class Line(LineBase):
    """
    The line vector graphics class.

    References
    ----------
    - Graphics draw_line interface document
        - https://bit.ly/3ey4pYe
    - Graphics draw_dotted_line interface document
        - https://bit.ly/3ig7Tzy
    - Graphics draw_dashed_line interface document
        - https://bit.ly/3ewoMF8
    - Graphics draw_round_dotted_line interface document
        - https://bit.ly/3ri985m
    - Graphics draw_dash_dotted_line interface document
        - https://bit.ly/3wKRtUZ
    """

    _start_point: 'point2d.Point2D'
    _end_point: 'point2d.Point2D'

    def __init__(
            self, *, parent: 'graphics.Graphics',
            start_point: 'point2d.Point2D',
            end_point: 'point2d.Point2D') -> None:
        """
        Create a line vector graphics.

        Parameters
        ----------
        parent : Graphics
            Graphics instance to link this graphic.
        start_point : Points2D
            Line start point.
        end_point : Points2D
            Line end point.

        References
        ----------
        - Graphics draw_line interface document
            - https://bit.ly/3ey4pYe
        - Graphics draw_dotted_line interface document
            - https://bit.ly/3ig7Tzy
        - Graphics draw_dashed_line interface document
            - https://bit.ly/3ewoMF8
        - Graphics draw_round_dotted_line interface document
            - https://bit.ly/3ri985m
        - Graphics draw_dash_dotted_line interface document
            - https://bit.ly/3wKRtUZ
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='__init__', locals_=locals(),
                module_name=__name__, class_=Line):
            from apysc._display.graphics import Graphics
            from apysc._expression import expression_variables_util
            from apysc._expression import var_names
            from apysc._validation import geom_validation
            geom_validation.validate_point_2d_type(point=start_point)
            geom_validation.validate_point_2d_type(point=end_point)
            parent_graphics: Graphics = parent
            variable_name: str = expression_variables_util.\
                get_next_variable_name(type_name=var_names.LINE)
            super(Line, self).__init__(
                parent=parent, x=0, y=0, variable_name=variable_name)
            self._start_point = start_point
            self._end_point = end_point
            self._set_initial_basic_values(parent=parent)
            self._append_constructor_expression()
            self._set_line_setting_if_not_none_value_exists(
                parent_graphics=parent_graphics)
            self._set_overflow_visible_setting()

    def _append_constructor_expression(self) -> None:
        """
        Append a constructor expression.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self._append_constructor_expression,
                locals_=locals(),
                module_name=__name__, class_=Line):
            from apysc._display.stage import get_stage_variable_name
            stage_variable_name: str = get_stage_variable_name()
            points_str: str = self._make_points_expression()
            expression: str = (
                f'var {self.variable_name} = {stage_variable_name}'
                f'\n  .line({points_str})'
                '\n  .attr({'
            )
            expression = self._append_basic_vals_expression(
                expression=expression,
                indent_num=2)
            expression += '\n  });'
            ap.append_js_expression(expression=expression)

    def _make_points_expression(self) -> str:
        """
        Make line start and end expression str.

        Returns
        -------
        expression : str
            Each points expression.
        """
        import apysc as ap
        start_point: ap.Point2D = self._start_point
        end_point: ap.Point2D = self._end_point
        expression: str = (
            f'{start_point.x.variable_name}, '
            f'{start_point.y.variable_name}, '
            f'{end_point.x.variable_name}, '
            f'{end_point.y.variable_name}'
        )
        return expression

    def __repr__(self) -> str:
        """
        Get a string representation of this instance (for the sake of
        debugging).

        Returns
        -------
        repr_str : str
            Type name and variable name will be set
            (e.g., `Line('<variable_name>')`).
        """
        repr_str: str = f"Line('{self.variable_name}')"
        return repr_str



"""Base class for each animation interface.
"""

from apysc._animation.animation_finish_interface import \
    AnimationFinishInterface
from apysc._animation.animation_pause_interface import AnimationPauseInterface
from apysc._animation.animation_play_interface import AnimationPlayInterface
from apysc._animation.animation_reset_interface import AnimationResetInterface
from apysc._animation.animation_reverse_interface import \
    AnimationReverseInterface
from apysc._animation.animation_time_interface import AnimationTimeInterface


class AnimationInterfaceBase(
        AnimationPauseInterface, AnimationPlayInterface,
        AnimationResetInterface, AnimationFinishInterface,
        AnimationReverseInterface, AnimationTimeInterface):
    """Base class for each animation interface.
    """



"""Class implementation for the continue.
"""


class Continue:
    """
    The loop continue expression class.

    Notes
    -----
    This class can be instantiated in the with loop statement,
    for example, after the `with ap.For(...):` statement.

    References
    ----------
    - Continue document
        - https://simon-ritchie.github.io/apysc/continue.html
    """

    def __init__(self) -> None:
        """
        The loop continue expression class.

        Notes
        -----
        This class can be instantiated in the with loop statement,
        for example, after the `with ap.For(...):` statement.

        References
        ----------
        - Continue document
            - https://simon-ritchie.github.io/apysc/continue.html

        Examples
        --------
        >>> import apysc as ap
        >>> arr: ap.Array = ap.Array(range(3))
        >>> with ap.For(arr) as i:
        ...     with ap.If(i == 1):
        ...         _ = ap.Continue()
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='__init__', locals_=locals(),
                module_name=__name__, class_=Continue):
            from apysc._loop import loop_count
            current_loop_count: int = loop_count.get_current_loop_count()
            if current_loop_count == 0:
                err_msg: str = (
                    'The `Continue` class can be instantiated in the with '
                    'loop statement, for example, after the '
                    '`with ap.For(...):` statement.'
                )
                raise Exception(err_msg)
            ap.append_js_expression(expression='continue;')



"""Implementations of Circle class.
"""

from typing import Union

from apysc._display import graphics
from apysc._display.cx_interface import CxInterface
from apysc._display.cy_interface import CyInterface
from apysc._display.line_base import LineBase
from apysc._display.radius_interface import RadiusInterface
from apysc._type.int import Int


class Circle(  # type: ignore
        CxInterface, CyInterface, LineBase, RadiusInterface):
    """
    The circle vector graphics class.

    References
    ----------
    - Graphics draw_circle interface document
        - https://bit.ly/36Sgmn3
    """

    def __init__(
            self,
            *,
            parent: 'graphics.Graphics',
            x: Union[int, Int],
            y: Union[int, Int],
            radius: Union[int, Int]) -> None:
        """
        Create a circle vector graphics.

        Parameters
        ----------
        parent : Graphics
            Graphics instance to link this graphics.
        x : int or Int
            X-coordinate of the circle center.
        y : int or Int
            Y-coordinate of the circle center.
        radius : int or Int
            Circle radius.

        References
        ----------
        - Graphics draw_circle interface document
            - https://bit.ly/36Sgmn3
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='__init__', locals_=locals(),
                module_name=__name__, class_=Circle):
            from apysc._expression import expression_variables_util
            from apysc._expression import var_names
            from apysc._validation import size_validation
            variable_name: str = expression_variables_util.\
                get_next_variable_name(type_name=var_names.CIRCLE)
            super(Circle, self).__init__(
                parent=parent, x=0, y=0, variable_name=variable_name)
            size_validation.validate_size_is_gt_zero(size=radius)
            self._radius = self._get_converted_radius_int(radius=radius)
            self._set_initial_basic_values(parent=parent)
            self._append_constructor_expression()
            self._set_center_coordinates(x=x, y=y)
            self._set_line_setting_if_not_none_value_exists(
                parent_graphics=parent)
            self._set_overflow_visible_setting()

            self._append_applying_new_attr_val_exp(
                new_attr=self._radius, attr_name='radius')
            self._append_attr_to_linking_stack(
                attr=self._radius, attr_name='radius')

    def _set_center_coordinates(
            self,
            *,
            x: Union[int, Int],
            y: Union[int, Int]) -> None:
        """
        Set a center x-coordinate and a center y-coordinate.

        Parameters
        ----------
        x : int or Int
            X-coordinate of the circle center.
        y : int or Int
            Y-coordinate of the circle center.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self._set_center_coordinates, locals_=locals(),
                module_name=__name__, class_=Circle):
            from apysc._converter.to_apysc_val_from_builtin import \
                get_copied_int_from_builtin_val
            self.x = get_copied_int_from_builtin_val(integer=x)
            self.y = get_copied_int_from_builtin_val(integer=y)

    def _append_constructor_expression(self) -> None:
        """
        Append a construcor expression.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self._append_constructor_expression,
                locals_=locals(),
                module_name=__name__, class_=Circle):
            from apysc._display.stage import get_stage_variable_name
            from apysc._type import value_util
            stage_variable_name: str = get_stage_variable_name()
            radius_str: str = value_util.get_value_str_for_expression(
                value=self._radius)
            expression: str = (
                f'var {self.variable_name} = {stage_variable_name}'
                f'\n  .circle({radius_str} * 2)'
                '\n  .attr({'
            )
            expression = self._append_basic_vals_expression(
                expression=expression, indent_num=2)
            expression += '\n  });'
            ap.append_js_expression(expression=expression)

    def __repr__(self) -> str:
        """
        Get a string representation of this instance (for the sake of
        debugging).

        Returns
        -------
        repr_str : str
            Type name and variable name will be set
            (e.g., `Circle('<variable_name>')`).
        """
        repr_str: str = f"Circle('{self.variable_name}')"
        return repr_str



"""Validation implementations for the handler's options value.
"""

from typing import Any


def validate_options_type(options: Any) -> None:
    """
    Validate a specified options type.

    Parameters
    ----------
    options : Any
        Target options value.

    Raises
    ------
    TypeError
        If a specified options type is not the dictionary or None.
    """
    if options is None:
        return
    if isinstance(options, dict):
        return
    raise TypeError(
        f"Handler's options argument must be a dictionary: {type(options)}"
        f'\n{options}')



"""Implementations for Graphics class.
"""

from typing import List
from typing import Optional
from typing import Union

from apysc._display import circle as _circle
from apysc._display import ellipse as _ellipse
from apysc._display import line as _line
from apysc._display import path as _path
from apysc._display import polygon as _polyg
from apysc._display import polyline as _polyline
from apysc._display import sprite
from apysc._display.begin_fill_interface import BeginFillInterface
from apysc._display.child_interface import ChildInterface
from apysc._display.display_object import DisplayObject
from apysc._display.graphics_clear_interface import GraphicsClearInterface
from apysc._display.line_style_interface import LineStyleInterface
from apysc._display.rectangle import Rectangle
from apysc._geom.path_data_base import PathDataBase
from apysc._geom.point2d import Point2D
from apysc._type.array import Array
from apysc._type.int import Int
from apysc._type.variable_name_interface import VariableNameInterface


class Graphics(
        DisplayObject,
        BeginFillInterface, LineStyleInterface, VariableNameInterface,
        GraphicsClearInterface, ChildInterface):
    """
    Create a object that has each vector graphics interface.

    References
    ----------
    - Graphics document
        - https://simon-ritchie.github.io/apysc/graphics.html
    """

    _current_line: Optional['_polyline.Polyline'] = None

    def __init__(
            self, *, parent: 'sprite.Sprite',
            variable_name: Optional[str] = None) -> None:
        """
        Create a object that has each vector graphics interface.

        Parameters
        ----------
        parent : Sprite
            This instance's parent instance.
        variable_name : str or None, default None
            Variable name to set. Specified only when subclass
            instantiation.

        References
        ----------
        - Graphics document
            - https://simon-ritchie.github.io/apysc/graphics.html
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='__init__', locals_=locals(),
                module_name=__name__, class_=Graphics):
            from apysc._expression import expression_variables_util
            from apysc._expression import var_names
            from apysc._validation import display_validation

            display_validation.validate_sprite(sprite=parent)
            self.parent_sprite: ap.Sprite = parent
            if variable_name is None:
                variable_name = expression_variables_util.\
                    get_next_variable_name(type_name=var_names.GRAPHICS)
            super(Graphics, self).__init__(variable_name=variable_name)
            self._fill_color = ap.String('')
            self._fill_alpha = ap.Number(1.0)
            self._line_color = ap.String('')
            self._line_alpha = ap.Number(1.0)
            self._line_thickness = ap.Int(1.0)
            self._children = ap.Array([])
            self._append_constructor_expression()
            self.parent_sprite.add_child(self)
            self._set_overflow_visible_setting()

    def _append_constructor_expression(self) -> None:
        """
        Append constructor expression.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self._append_constructor_expression,
                locals_=locals(),
                module_name=__name__, class_=Graphics):
            stage_name: str = self.parent_sprite.stage.variable_name
            expression: str = (
                f'var {self.variable_name} = {stage_name}.nested();'
            )
            ap.append_js_expression(expression=expression)

    def draw_rect(
            self, x: Union[int, Int],
            y: Union[int, Int],
            width: Union[int, Int],
            height: Union[int, Int]) -> Rectangle:
        """
        Draw a rectangle vector graphics.

        Parameters
        ----------
        x : int or Int
            X position to start drawing.
        y : int or Int
            Y position to start drawing.
        width : int or Int
            Rectangle width.
        height : int or Int
            Rectangle height.

        Returns
        -------
        rectangle : Rectangle
            Created rectangle.

        References
        ----------
        - Graphics draw_rect interface document
            - https://bit.ly/3zbSG9o
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self.draw_rect, locals_=locals(),
                module_name=__name__, class_=Graphics):
            rectangle: Rectangle = Rectangle(
                parent=self, x=x, y=y, width=width, height=height)
            self.add_child(child=rectangle)
            return rectangle

    def draw_round_rect(
            self, x: Union[int, Int],
            y: Union[int, Int],
            width: Union[int, Int],
            height: Union[int, Int],
            ellipse_width: Union[int, Int],
            ellipse_height: Union[int, Int]) -> Rectangle:
        """
        Draw a rounded rectangle vector graphics.

        Parameters
        ----------
        x : int or Int
            X-coordinate to start drawing.
        y : int or Int
            Y-coordinate to start drawing.
        width : int or Int
            Rectangle width.
        height : int or Int
            Rectangle height.
        ellipse_width : int or Int
            Ellipse width of the rectangle corner.
        ellipse_height : int or Int
            Ellipse height of the rectangle corner.

        Returns
        -------
        rectangle : Rectangle
            Created rectangle.

        References
        ----------
        - Graphics draw_round_rect interface document
            - https://bit.ly/2ThGcxJ
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self.draw_round_rect, locals_=locals(),
                module_name=__name__, class_=Graphics):
            rectangle: Rectangle = Rectangle(
                parent=self, x=x, y=y, width=width, height=height)
            if isinstance(ellipse_width, int):
                ellipse_width = ap.Int(ellipse_width)
            if isinstance(ellipse_height, int):
                ellipse_height = ap.Int(ellipse_height)
            rectangle.ellipse_width = ellipse_width
            rectangle.ellipse_height = ellipse_height
            self.add_child(child=rectangle)
            return rectangle

    def draw_circle(
            self,
            x: Union[int, Int],
            y: Union[int, Int],
            radius: Union[int, Int]) -> '_circle.Circle':
        """
        Draw a circle vector graphics.

        Parameters
        ----------
        x : int or Int
            X-coordinate of the circle center.
        y : int or Int
            Y-coordinate of the circle center.
        radius : int or Int
            Circle radius.

        Returns
        -------
        circle : Circle
            Created circle graphics instance.

        References
        ----------
        - Graphics draw_circle interface document
            - https://bit.ly/3it1q4y
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self.draw_circle, locals_=locals(),
                module_name=__name__, class_=Graphics):
            circle: _circle.Circle = _circle.Circle(
                parent=self, x=x, y=y, radius=radius)
            self.add_child(child=circle)
            return circle

    def draw_ellipse(
            self,
            x: Union[int, Int],
            y: Union[int, Int],
            width: Union[int, Int],
            height: Union[int, Int]) -> '_ellipse.Ellipse':
        """
        Draw a ellipse vector graphics.

        Parameters
        ----------
        x : int or Int
            X-coordinate of the ellipse center.
        y : int or Int
            Y-coordinate of the ellipse center.
        width : int or Int
            Ellipse width.
        height : int or Int
            Ellipse height.

        Returns
        -------
        ellipse : Ellipse
            Created ellipse graphics instance.

        References
        ----------
        - Graphics draw_ellipse interface
            - https://bit.ly/3xPVicP
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self.draw_ellipse, locals_=locals(),
                module_name=__name__, class_=Graphics):
            ellipse: _ellipse.Ellipse = _ellipse.Ellipse(
                parent=self, x=x, y=y, width=width, height=height)
            self.add_child(child=ellipse)
            return ellipse

    def line_to(
            self, x: Union[int, Int],
            y: Union[int, Int]) -> '_polyline.Polyline':
        """
        Draw a line from previous point to specified point (initial
        point is x = 0, y = 0).

        Parameters
        ----------
        x : int or Int
            X destination point to draw line.
        y : int or Int
            Y destination point to draw line.

        Returns
        -------
        line : Polyline
            Line graphics instance.

        References
        ----------
        - Graphics move_to and line_to interfaces document
            - https://bit.ly/3eybhEP
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self.line_to, locals_=locals(),
                module_name=__name__, class_=Graphics):
            if self._current_line is None:
                self._current_line = _polyline.Polyline(
                    parent=self,
                    points=ap.Array([Point2D(x=0, y=0), Point2D(x=x, y=y)]))
                self.add_child(self._current_line)
            else:
                self._current_line.append_line_point(x=x, y=y)
            return self._current_line

    def move_to(
            self,
            x: Union[int, Int],
            y: Union[int, Int]) -> '_polyline.Polyline':
        """
        Move a line position to specified point.

        Parameters
        ----------
        x : int or Int
            X destination point to move to.
        y : int or Int
            Y destination point to move to.

        Returns
        -------
        line : Polyline
            Line graphics instance.

        References
        ----------
        - Graphics move_to and line_to interfaces document
            - https://bit.ly/3eybhEP
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self.move_to, locals_=locals(),
                module_name=__name__, class_=Graphics):
            self._current_line = _polyline.Polyline(
                parent=self, points=ap.Array([Point2D(x=x, y=y)]))
            return self._current_line

    def _reset_each_line_settings(self) -> None:
        """
        Reset each line settings (e.g., LineDotSetting, LineDashSetting,
        and so on).

        Notes
        -----
        expression will not be appended.
        """
        self._line_dot_setting = None
        self._line_dash_setting = None
        self._line_round_dot_setting = None
        self._line_dash_dot_setting = None

    def draw_line(
            self,
            x_start: Union[int, Int],
            y_start: Union[int, Int],
            x_end: Union[int, Int],
            y_end: Union[int, Int]) -> '_line.Line':
        """
        Draw a normal line vector graphics.

        Notes
        -----
        - This interface will ignore line settings, like a
            LineDotSetting, LineDashSetting, and so on.

        Parameters
        ----------
        x_start : int or Int
            Line start x-coordinate.
        y_start : int or Int
            Line start y-coordinate.
        x_end : int or Int
            Line end x-coordinate.
        y_end : int or Int
            Line end y-coordinate.

        Returns
        -------
        line : Line
            Created line graphics instance.

        References
        ----------
        - Graphics draw_line interface document
            - https://bit.ly/3ey4pYe
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self.draw_line, locals_=locals(),
                module_name=__name__, class_=Graphics):
            snapshot_name: str = self._get_next_snapshot_name()
            self._run_all_make_snapshot_methods(snapshot_name=snapshot_name)
            self._reset_each_line_settings()
            line: _line.Line = _line.Line(
                parent=self,
                start_point=Point2D(x=x_start, y=y_start),
                end_point=Point2D(x=x_end, y=y_end))
            self._run_all_revert_methods(snapshot_name=snapshot_name)
            self.add_child(child=line)
            return line

    def draw_dotted_line(
            self,
            x_start: Union[int, Int],
            y_start: Union[int, Int],
            x_end: Union[int, Int],
            y_end: Union[int, Int],
            dot_size: Union[int, Int]) -> '_line.Line':
        """
        Draw a dotted line vector graphics.

        Notes
        -----
        - This interface will ignore line settings, line a
            LineDashSetting, except LineDotSetting.

        Parameters
        ----------
        x_start : int or Int
            Line start x-coordinate.
        y_start : int or Int
            Line start y-coordinate.
        x_end : int or Int
            Line end x-coordinate.
        y_end : int or Int
            Line end y-coordinate.
        dot_size : int or Int
            Dot size.

        Returns
        -------
        line : Line
            Created line graphics instance.

        References
        ----------
        - Graphics draw_dotted_line interface document
            - https://bit.ly/3ig7Tzy
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self.draw_dotted_line, locals_=locals(),
                module_name=__name__, class_=Graphics):
            snapshot_name: str = self._get_next_snapshot_name()
            self._run_all_make_snapshot_methods(snapshot_name=snapshot_name)
            self._reset_each_line_settings()
            self._line_dot_setting = ap.LineDotSetting(dot_size=dot_size)
            line: _line.Line = _line.Line(
                parent=self,
                start_point=Point2D(x=x_start, y=y_start),
                end_point=Point2D(x=x_end, y=y_end))
            self._run_all_revert_methods(snapshot_name=snapshot_name)
            self.add_child(child=line)
            return line

    def draw_dashed_line(
            self,
            x_start: Union[int, Int],
            y_start: Union[int, Int],
            x_end: Union[int, Int],
            y_end: Union[int, Int],
            dash_size: Union[int, Int],
            space_size: Union[int, Int]) -> '_line.Line':
        """
        Draw a dashed line vector graphics.

        Notes
        -----
        - This interface will ignore line settings, like a
            LineDotSetting, except LineDashSetting.

        Parameters
        ----------
        x_start : int or Int
            Line start x-coordinate.
        y_start : int or Int
            Line start y-coordinate.
        x_end : int or Int
            Line end x-coordinate.
        y_end : int or Int
            Line end y-coordinate.
        dash_size : int or Int
            Dash size.
        space_size : int or Int
            Blank space size between dashes.

        Returns
        -------
        line : Line
            Created line graphics instance.

        References
        ----------
        - Graphics draw_dashed_line interface document
            - https://bit.ly/3ewoMF8
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self.draw_dashed_line, locals_=locals(),
                module_name=__name__, class_=Graphics):
            snapshot_name: str = self._get_next_snapshot_name()
            self._run_all_make_snapshot_methods(snapshot_name=snapshot_name)
            self._reset_each_line_settings()
            self._line_dash_setting = ap.LineDashSetting(
                dash_size=dash_size, space_size=space_size)
            line: _line.Line = _line.Line(
                parent=self,
                start_point=Point2D(x=x_start, y=y_start),
                end_point=Point2D(x=x_end, y=y_end))
            self._run_all_revert_methods(snapshot_name=snapshot_name)
            self.add_child(child=line)
            return line

    def draw_round_dotted_line(
            self,
            x_start: Union[int, Int],
            y_start: Union[int, Int],
            x_end: Union[int, Int],
            y_end: Union[int, Int],
            round_size: Union[int, Int],
            space_size: Union[int, Int]) -> '_line.Line':
        """
        Draw a round dotted line vector graphics.

        Notes
        -----
        - This interface will ignore line settings, like a
            LineDotSetting, except LineRoundDotSetting.

        Parameters
        ----------
        x_start : int or Int
            Line start x-coordinate.
        y_start : int or Int
            Line start y-coordinate.
        x_end : int or Int
            Line end x-coordinate.
        y_end : int or Int
            Line end y-coordinate.
        round_size : int or Int
            Dot round size.
        space_size : int or Int
            Blank space size between dots.

        Returns
        -------
        line : Line
            Created line graphics instance.

        References
        ----------
        - Graphics draw_round_dotted_line interface document
            - https://bit.ly/3ri985m
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self.draw_round_dotted_line, locals_=locals(),
                module_name=__name__, class_=Graphics):
            snapshot_name: str = self._get_next_snapshot_name()
            self._run_all_make_snapshot_methods(snapshot_name=snapshot_name)
            self._reset_each_line_settings()
            self._line_round_dot_setting = ap.LineRoundDotSetting(
                round_size=round_size, space_size=space_size)
            line: _line.Line = _line.Line(
                parent=self,
                start_point=Point2D(x=x_start, y=y_start),
                end_point=Point2D(x=x_end, y=y_end))
            self._run_all_revert_methods(snapshot_name=snapshot_name)
            self.add_child(child=line)
            return line

    def draw_dash_dotted_line(
            self,
            x_start: Union[int, Int],
            y_start: Union[int, Int],
            x_end: Union[int, Int],
            y_end: Union[int, Int],
            dot_size: Union[int, Int],
            dash_size: Union[int, Int],
            space_size: Union[int, Int]) -> '_line.Line':
        """
        Draw a dash dotted (1-dot chain) line vector graphics.

        Parameters
        ----------
        x_start : int or Int
            Line start x-coordinate.
        y_start : int or Int
            Line start y-coordinate.
        x_end : int or Int
            Line end x-coordinate.
        y_end : int or Int
            Line end y-coordinate.
        dot_size : int or Int
            Dot size.
        dash_size : int or Int
            Dash size.
        space_size : int or Int
            Blank space size between dots and dashes.

        Returns
        -------
        line : Line
            Created line graphics instance.

        References
        ----------
        - Graphics draw_dash_dotted_line interface document
            - https://bit.ly/3wKRtUZ
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self.draw_dash_dotted_line, locals_=locals(),
                module_name=__name__, class_=Graphics):
            snapshot_name: str = self._get_next_snapshot_name()
            self._run_all_make_snapshot_methods(snapshot_name=snapshot_name)
            self._reset_each_line_settings()
            self._line_dash_dot_setting = ap.LineDashDotSetting(
                dot_size=dot_size,
                dash_size=dash_size,
                space_size=space_size)
            line: _line.Line = _line.Line(
                parent=self,
                start_point=Point2D(x=x_start, y=y_start),
                end_point=Point2D(x=x_end, y=y_end))
            self._run_all_revert_methods(snapshot_name=snapshot_name)
            self.add_child(child=line)
            return line

    def draw_polygon(
            self,
            points: Union[List[Point2D], Array[Point2D]]) -> '_polyg.Polygon':
        """
        Draw a polygon vector graphics. This is similar to Polyline
        class (created by move_to or line_to, or other interface),
        but unlike that, end point and start one will be connected.

        Parameters
        ----------
        points : list of Point2D or Array.
            Polygon vertex points.

        Returns
        -------
        polygon : Polygon
            Created polygon graphics instance.

        References
        ----------
        - Graphics draw_polygon interface document
            - https://bit.ly/3wHVZUk
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self.draw_polygon, locals_=locals(),
                module_name=__name__, class_=Graphics):
            if isinstance(points, list):
                points = ap.Array(points)
            polygon: _polyg.Polygon = _polyg.Polygon(
                parent=self, points=points)
            self.add_child(polygon)
            return polygon

    def draw_path(
            self, path_data_list: List[PathDataBase]) -> '_path.Path':
        """
        Draw a path vector graphics.

        Parameters
        ----------
        path_data_list : list of PathDataBase
            Target path data settings, such as the ap.PathData.MoveTo.

        Returns
        -------
        path : Path
            Created path graphics instance.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self.draw_path, locals_=locals(),
                module_name=__name__, class_=Graphics):
            path: _path.Path = _path.Path(
                parent=self, path_data_list=path_data_list)
            self.add_child(path)
            return path

    def __repr__(self) -> str:
        """
        Get a string representation of this instance (for the sake of
        debugging).

        Returns
        -------
        repr_str : str
            Type name and variable name will be set
            (e.g., `Graphics('<variable_name>')`).
        """
        repr_str: str = f"Graphics('{self.variable_name}')"
        return repr_str



"""Class implementation for the center y-coordinate interface.
"""

from typing import Dict

from apysc._animation.animation_cy_interface import AnimationCyInterface
from apysc._type.attr_linking_interface import AttrLinkingInterface
from apysc._type.int import Int
from apysc._type.revert_interface import RevertInterface


class CyInterface(
        AnimationCyInterface, RevertInterface, AttrLinkingInterface):

    _cy: Int

    def _initialize_cy_if_not_initialized(self) -> None:
        """
        Initialize _cy attribute if it hasn't been initialized yet.
        """
        if hasattr(self, '_cy'):
            return
        self._cy = Int(0)

        self._append_cy_attr_linking_setting()

    def _append_cy_attr_linking_setting(self) -> None:
        """
        Append a cy attribute linking setting.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self._append_cy_attr_linking_setting,
                locals_=locals(),
                module_name=__name__, class_=CyInterface):
            self._append_applying_new_attr_val_exp(
                new_attr=self._cy, attr_name='cy')
            self._append_attr_to_linking_stack(attr=self._cy, attr_name='cy')

    @property
    def y(self) -> Int:
        """
        Get a center y-coordinate.

        Returns
        -------
        y : Int
            Center y-coordinate.

        References
        ----------
        - Display object x and y interfaces document
            - https://bit.ly/3q7GFQM

        Examples
        --------
        >>> import apysc as ap
        >>> stage: ap.Stage = ap.Stage()
        >>> sprite: ap.Sprite = ap.Sprite()
        >>> sprite.graphics.begin_fill(color='#0af', alpha=0.5)
        >>> circle: ap.Circle = sprite.graphics.draw_circle(
        ...     x=100, y=100, radius=50)
        >>> circle.y = ap.Int(120)
        >>> circle.y
        Int(120)
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='y', locals_=locals(),
                module_name=__name__, class_=CyInterface):
            from apysc._type import value_util
            self._initialize_cy_if_not_initialized()
            y: ap.Int = value_util.get_copy(value=self._cy)
            return y

    @y.setter
    def y(self, value: Int) -> None:
        """
        Update a center y-coordinate.

        Parameters
        ----------
        value : int or Int
            Center y-coordinate value.

        References
        ----------
        - Display object x and y interfaces document
            - https://bit.ly/3q7GFQM
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='y', locals_=locals(),
                module_name=__name__, class_=CyInterface):
            from apysc._validation import number_validation
            number_validation.validate_integer(integer=value)
            if not isinstance(value, ap.Int):
                value = ap.Int(value)
            self._cy = value
            self._cy._append_incremental_calc_substitution_expression()
            self._append_cy_update_expression()

            self._append_cy_attr_linking_setting()

    def _append_cy_update_expression(self) -> None:
        """
        Append cy position updating expression.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self._append_cy_update_expression, locals_=locals(),
                module_name=__name__, class_=CyInterface):
            from apysc._type import value_util
            self._initialize_cy_if_not_initialized()
            value_str: str = value_util.get_value_str_for_expression(
                value=self._cy)
            expression: str = (
                f'{self.variable_name}.cy({value_str});'
            )
            ap.append_js_expression(expression=expression)

    _cy_snapshots: Dict[str, int]

    def _make_snapshot(self, *, snapshot_name: str) -> None:
        """
        Make a value's snapshot.

        Parameters
        ----------
        snapshot_name : str
            Target snapshot name.
        """
        self._initialize_cy_if_not_initialized()
        self._set_single_snapshot_val_to_dict(
            dict_name='_cy_snapshots',
            value=int(self._cy._value), snapshot_name=snapshot_name)

    def _revert(self, *, snapshot_name: str) -> None:
        """
        Revert a value if snapshot exists.

        Parameters
        ----------
        snapshot_name : str
            Target snapshot name.
        """
        if not self._snapshot_exists(snapshot_name=snapshot_name):
            return
        self._cy._value = self._cy_snapshots[snapshot_name]



"""Each interface to get an apysc value from a Python built-in one.
"""

from typing import Union

from apysc._type.boolean import Boolean
from apysc._type.int import Int
from apysc._type.number import Number
from apysc._type.string import String


def get_copied_int_from_builtin_val(
        *, integer: Union[int, Int]) -> Int:
    """
    Get a copied Int value from a Python built-in int.

    Parameters
    ----------
    integer : int or Int
        Target integer value.

    Returns
    -------
    copied : Int
        Copied Int value.
    """
    import apysc as ap
    with ap.DebugInfo(
            callable_=get_copied_int_from_builtin_val, locals_=locals(),
            module_name=__name__):
        if isinstance(integer, int):
            copied: ap.Int = ap.Int(integer)
        else:
            copied = integer._copy()
        return copied


def get_copied_number_from_builtin_val(
        *, float_or_num: Union[float, Number]) -> Number:
    """
    Get a copied number value from a Python built-in float.

    Parameters
    ----------
    float_or_num : float or Number
        Target float (or Number) value.

    Returns
    -------
    num : Number
        Copied Number value.
    """
    import apysc as ap
    with ap.DebugInfo(
            callable_=get_copied_number_from_builtin_val, locals_=locals(),
            module_name=__name__):
        if isinstance(float_or_num, float):
            copied: ap.Number = ap.Number(float_or_num)
        else:
            copied = float_or_num._copy()
        return copied


def get_copied_string_from_builtin_val(
        *, string: Union[str, String]) -> String:
    """
    Get a copied String value from a Python built-in str.

    Parameters
    ----------
    string : str or String
        Target string value.

    Returns
    -------
    copied : String
        Copied String value.
    """
    import apysc as ap
    with ap.DebugInfo(
            callable_=get_copied_string_from_builtin_val, locals_=locals(),
            module_name=__name__):
        if isinstance(string, str):
            copied: ap.String = ap.String(string)
        else:
            copied = string._copy()
        return copied


def get_copied_boolean_from_builtin_val(
        *, bool_val: Union[bool, Boolean]) -> Boolean:
    """
    Get a copied Boolean value from a Python built-in bool.

    Parameters
    ----------
    bool_val : bool
        Target bool value.

    Returns
    -------
    copied : Boolean
        Copied Boolean value.
    """
    import apysc as ap
    with ap.DebugInfo(
            callable_=get_copied_boolean_from_builtin_val, locals_=locals(),
            module_name=__name__):
        if isinstance(bool_val, bool):
            copied: ap.Boolean = ap.Boolean(bool_val)
        else:
            copied = bool_val._copy()
        return copied
